from typing import Union
from typing import Callable
from typing import Optional
from typing import List

import ast
import re
import functools

from loguru import logger
from lark import Lark, ParseTree

import datamol as dm

from medchem.catalogs import NamedCatalogs
from medchem.catalogs import list_named_catalogs
from medchem.groups import list_functional_group_names
from medchem.groups import get_functional_group_map
from medchem.rules import list_descriptors
from medchem.rules import RuleFilters
from medchem.rules import basic_rules
from medchem.utils.loader import get_grammar

from ._parser import QueryParser


class QueryOperator:
    """A class to hold all the operators that can be used in queries"""

    AVAILABLES_PROPERTIES = list_descriptors()  # list of available molecular properties
    """Default list of available properties in medchem's query system"""

    AVAILABLE_CATALOGS = list_named_catalogs()  # list of available catalogs
    """Default list of available catalogs in medchem's query system"""

    AVAILABLE_RULES = RuleFilters.list_available_rules_names()  # list of available rules
    """Default list of available rules in medchem's query system"""

    AVAILABLE_FUNCTIONAL_GROUPS = list_functional_group_names()  # list of available functional groups
    """Default list of available functional groups in medchem's query system"""

    @staticmethod
    def hassubstructure(
        mol: Union[dm.Mol, str],
        query: str,
        is_smarts: bool = False,
        operator: Optional[str] = "min",
        limit: int = 1,
    ):
        """Check if a molecule has the substructure provided by a query

        Args:
            mol: input molecule
            query: input smarts query
            is_smarts: whether this is a smarts query or not
            operator: one of min or max to specify the min or max limit
            limit: limit of substructures to be found

        Returns:
            has_substructure (bool): whether the query is a subgraph of the molecule
        """
        if is_smarts:
            _query = dm.from_smarts(query)
        else:
            _query = dm.to_mol(query)

        if limit is None:
            limit = 1

        if operator is None:
            operator = "min"

        mol = dm.to_mol(mol)

        if mol is None:
            raise ValueError("Molecule is None")

        matches = mol.GetSubstructMatches(_query, uniquify=True)
        coeff = -1 if operator == "min" else 1

        return len(matches) * coeff <= limit * coeff

    @staticmethod
    def hassuperstructure(mol: Union[dm.Mol, str], query: str):
        """Check if a molecule has a superstructure defined by a query.
        Note that a superstructure cannot be a query (SMARTS)

        Args:
            mol: input molecule
            query: input smarts query

        Returns:
            has_superstructure (bool): whether the molecule is a subgraph of the query
        """
        _query = dm.to_mol(query)
        mol = dm.to_mol(mol)

        if mol is None:
            raise ValueError("Molecule is None")

        if _query is None:
            raise ValueError("Query is None")

        return _query.HasSubstructMatch(mol)

    @staticmethod
    def hasalert(mol: Union[dm.Mol, str], alert: str):
        """Check if a molecule match a named alert catalog.
        The alert catalog needs to be one supported by the medchem package.

        Args:
            mol: input molecule
            alert: named catalog to apply as filter on the molecule

        Returns:
            has_alert (bool): whether the molecule has a given alert
        """

        mol = dm.to_mol(mol)

        if mol is None:
            raise ValueError("Molecule is None")

        if isinstance(alert, str):
            if alert not in QueryOperator.AVAILABLE_CATALOGS:
                raise ValueError(
                    f"Alert {alert} is not supported. Available alerts are: {QueryOperator.AVAILABLE_CATALOGS}"
                )
            _alert = getattr(NamedCatalogs, alert, lambda: None)()

        else:
            _alert = alert

        if _alert is None:
            raise ValueError("Alert is None")

        return _alert.HasMatch(mol)

    @staticmethod
    def matchrule(mol: Union[dm.Mol, str], rule: str):
        """Check if a molecule match a druglikeness rule

        Args:
            mol: input molecule
            rule: druglikeness rule check on the molecule.

        Returns:
            match_rule (bool): whether the molecule match the given rule
        """

        mol = dm.to_mol(mol)

        if mol is None:
            raise ValueError("Molecule is None")

        if isinstance(rule, str):
            if rule not in QueryOperator.AVAILABLE_RULES:
                raise ValueError(
                    f"Rule {rule} is not supported. Available rules are: {QueryOperator.AVAILABLE_RULES}"
                )
            _rule = getattr(basic_rules, rule, None)
        else:
            _rule = rule

        if _rule is None:
            raise ValueError("Rule is None")

        return _rule(mol)

    @staticmethod
    def hasgroup(mol: Union[dm.Mol, str], group: str):
        """Check if a molecule has a specific functional group.

        Args:
            mol: input molecule
            group: functional group to check on the molecule.

        Returns:
            has_group (bool): whether the molecule has the given functional group
        """

        mol = dm.to_mol(mol)
        if group not in QueryOperator.AVAILABLE_FUNCTIONAL_GROUPS:
            raise ValueError(
                f"Functional Group {group} is not supported. Available functional group are: {QueryOperator.AVAILABLE_FUNCTIONAL_GROUPS}"
            )
        group_smarts = get_functional_group_map()[group]
        return QueryOperator.hassubstructure(mol, group_smarts, is_smarts=True)

    @staticmethod
    def hasprop(mol: Union[dm.Mol, str], prop: str, comparator: Callable, limit: float):
        """Check if a molecule has a property within a desired range

        Args:
            mol: input molecule
            prop: molecular property to apply as filter on the molecule
            comparator: operator function to apply to check whether the molecule property matches the expected value
            limit: limit value for determining whether the molecule property is within desired range

        Returns:
            has_property (bool): whether the molecule has a given property within a desired range
        """

        mol = dm.to_mol(mol)
        if isinstance(prop, str):
            if prop not in QueryOperator.AVAILABLES_PROPERTIES:
                raise ValueError(
                    f"Property {prop} is not supported. Available properties are: {QueryOperator.AVAILABLES_PROPERTIES}"
                )
            _prop = getattr(dm.descriptors, prop, None)

        else:
            _prop = prop

        if _prop is None:
            raise ValueError(f"Prop is None: '{prop}'")

        computed = _prop(mol)
        return comparator(computed, limit)

    @staticmethod
    def getprop(mol: Union[dm.Mol, str], prop: str):
        """Compute the molecular property of a molecule.
        This is an alternative to the hasprop function, that does not enforce any comparison.

        Args:
            mol: input molecule
            prop: molecular property to apply as filter on the molecule

        Returns:
            property (float):  computed property value
        """

        mol = dm.to_mol(mol)
        prop_fn = None
        if isinstance(prop, str):
            if prop not in QueryOperator.AVAILABLES_PROPERTIES:
                raise ValueError(
                    f"Property {prop} is not supported. Available properties are: {QueryOperator.AVAILABLES_PROPERTIES}"
                )
            prop_fn = getattr(dm.descriptors, prop, None)

        else:
            prop_fn = prop

        if prop_fn is None:
            raise ValueError("Prop is None")

        return prop_fn(mol)

    @staticmethod
    def like(
        mol: Union[dm.Mol, str],
        query: Union[dm.Mol, str],
        comparator: Callable[[float, float], bool],
        limit: float,
    ):
        """Check if a molecule is similar or distant enough from another molecule using tanimoto ECFP distance.

        Args:
            mol: input molecule
            query: input molecule to compare with
            comparator: operator function to apply to check whether the molecule property matches the expected value.
                Takes computed_similarity and `limit` as arguments and returns a boolean.
            limit: limit value for determining whether the molecule property is within desired range

        Returns:
            is_similar (bool): whether the molecule is similar or distant enough from the query
        """

        mol = dm.to_mol(mol)
        query = dm.to_mol(query)
        sim_val = 1 - dm.cdist([mol], [query])[0][0]
        return comparator(sim_val, limit)

    @staticmethod
    def similarity(
        mol: Union[dm.Mol, str],
        query: Union[dm.Mol, str],
    ):
        """Compute the ECFP tanimoto similarity between two molecules.
        This is an alternative to the like function, that does not enforce any comparison,
        and lets python handles the binary comparison operators.

        Args:
            mol: input molecule
            query: input query molecule to compute similarity against

        Returns:
            similarity (float): computed similarity value between mol and query
        """

        mol = dm.to_mol(mol)
        query = dm.to_mol(query)
        sim_val = 1 - dm.cdist([mol], [query])[0][0]
        return sim_val


class _NodeEvaluator:
    """Representation and evaluation of a single node in the grammar tree"""

    def __init__(self, node_expr: str):
        """Initialize a node evaluator

        Args:
            node_expr: node expression to evaluate

        Raises:
            ValueError: if the node expression is not valid
        """
        self.node_expr = node_expr
        self.node_fn = None
        if self.node_expr.startswith("fn("):
            # EN: revisit this with a regexp eventually for robustness
            node_expr = node_expr[3:-1]  # remove `fn(` and `)`
            node_arg_list = node_expr.split(", ")
            _fn = getattr(QueryOperator, node_arg_list[0], None)
            if _fn is None:
                raise ValueError("Unknown function {}".format(node_arg_list[0]))
            _kwargs = dict((k, ast.literal_eval(v)) for k, v in [x.split("=", 1) for x in node_arg_list[1:]])
            self.node_fn = functools.partial(_fn, **_kwargs)

    def __call__(self, *args, **kwargs):
        if self.node_fn is None:
            return self.node_expr
        return self.node_fn(*args, **kwargs)


class EvaluableQuery:
    """Parser of a query into a list of evaluable function nodes"""

    FN_PATTERN = re.compile("`(.*?)`")

    def __init__(self, parsed_query: Union[str, ParseTree], verbose: bool = False):
        """Constructor for query evaluation

        Args:
            parsed_query: query that has been parsed and transformed
            verbose: whether to print debug information
        """

        parsed_query = str(parsed_query)
        self.query_nodes = [_NodeEvaluator(x) for x in self.FN_PATTERN.split(parsed_query)]
        self.verbose = verbose

    def __call__(self, mol: Union[dm.Mol, str], exec: bool = True) -> str:
        """Evaluate a query on an input molecule

        Args:
            mol: input molecule
            exec: whether to interpret the resulting query or not

        Returns:
            query string or boolean value corresponding to the query result
        """
        query_eval = " ".join([f"{node(mol)}" for node in self.query_nodes])
        if self.verbose:
            logger.debug(query_eval)
        if exec:
            # EN: eval is not safe, but we are using it
            # because ast.literal_eval cannot parse some tree structures
            # and also because anything remaining here is sanitized or just boolean expression
            return eval(query_eval)
        return query_eval


class QueryFilter:
    """
    Query filtering system based on a custom query grammar
    """

    def __init__(self, query: str, grammar: Optional[str] = None, parser: str = "lalr"):
        """Constructor for query filtering system

        Args:
            query: input unparsed query
            grammar: path to grammar language to use. Defaults to None, which will use the default grammar.
            parser: which Lark language parser to use. Defaults to "lalr".
        """
        if parser not in ["earley", "lalr"]:
            raise AttributeError("parser must be either 'earley' or 'lalr'")
        # if existing file, then load and parse it
        self.grammar = get_grammar(grammar, as_string=True)
        self.query_parser = Lark(self.grammar, parser=parser)
        self.transformer = QueryParser()
        self._query_str = query
        self.query = self.transformer.transform(self.query_parser.parse(self._query_str))
        self._evaluable_query = EvaluableQuery(self.query)

    def __repr__(self) -> str:
        return self.query

    def __call__(
        self,
        mols: List[Union[str, dm.Mol]],
        scheduler: str = "processes",
        n_jobs: int = -1,
        progress: bool = True,
    ):
        """Call the internal chemical filter that has been build

        Args:
            mols: list of input molecules to filter
            n_jobs: whether to run jobs in parallel and number of jobs to consider.
            scheduler: joblib scheduler to use.
            progress: whether to show job progress.
        """
        tqdm_kwargs = {}
        if progress:
            tqdm_kwargs = {"desc": "Loading Mols", "leave": False}

        results = dm.parallelized(
            self._evaluable_query,
            mols,
            n_jobs=n_jobs,
            scheduler=scheduler,
            progress=progress,
            tqdm_kwargs=tqdm_kwargs,
        )
        return list(results)



from typing import Union
from typing import Callable
from typing import Optional
from typing import List
from typing import Sequence
from typing import Dict

import functools

import pandas as pd
import datamol as dm

from medchem.utils import loader

from ..utils.func import get_function_name
from . import basic_rules


def _compute_batch_props(mols):
    """Wrapper for making batch computing property under parallelization"""
    properties_fn = {
        "mw": dm.descriptors.mw,
        "n_rings": dm.descriptors.n_rings,
        "n_heavy_atoms": dm.descriptors.n_heavy_atoms,
        "n_rotatable_bonds": dm.descriptors.n_rotatable_bonds,
        "tpsa": dm.descriptors.tpsa,
        "clogp": dm.descriptors.clogp,
        "n_hba": dm.descriptors.n_hba,
        "n_hbd": dm.descriptors.n_hbd,
        "n_lipinski_hba": dm.descriptors.n_lipinski_hba,
        "n_lipinski_hbd": dm.descriptors.n_lipinski_hbd,
    }
    return dm.descriptors.compute_many_descriptors(mols, properties_fn, add_properties=False)


class RuleFilters:
    """
    Build a filter based on a compound phychem properties. For a list of default rules, use `RuleFilters.list_available_rules()`.
    Most of these rules have been collected from the litterature including https://fafdrugs4.rpbs.univ-paris-diderot.fr/descriptors.html
    """

    def __init__(
        self,
        rule_list: List[Union[str, Callable]],
        rule_list_names: Optional[List[Optional[str]]] = None,
    ):
        """Build a rule filtering object

        Args:
            rule_list: list of rules to apply. Either a callable that takes a molecule as input (with kwargs) or a string
                of the name of a pre-defined rule as defined in the basic_rules module
            rule_list_names: Name of the rules passed as inputs. Defaults to None.
        """
        self.rules = self._build_rules(rule_list, rule_list_names)

    def _build_rules(
        self,
        rule_list: List[Union[str, Callable]],
        rule_list_names: Optional[List[Optional[str]]] = None,
    ) -> Dict[str, Callable]:
        """Build the list of rules to apply."""

        if rule_list_names is not None and len(rule_list_names) != len(rule_list):
            raise ValueError("rule_list_names must be the same length as rule_list")

        if rule_list_names is None:
            _rule_list_names = [None] * len(rule_list)
        else:
            _rule_list_names = rule_list_names

            # If rule_list_names is provided, we check whether the list is unique
            if len(set(_rule_list_names)) != len(_rule_list_names):
                raise ValueError("rule_list_names must be unique")

        rules = {}
        for rule_name, rule_fn in zip(_rule_list_names, rule_list):
            if isinstance(rule_fn, str):
                rule = getattr(basic_rules, rule_fn, None)

                if rule is None:
                    raise ValueError(f"Rule {rule_fn} not found")

            elif callable(rule_fn):
                rule = rule_fn

            else:
                raise ValueError(f"Unsupported rule {rule_fn} of type {type(rule_fn)}!")

            if rule_name is None:
                if isinstance(rule_fn, str):
                    rule_name = rule_fn
                else:
                    rule_name = get_function_name(rule_fn)

            rules[rule_name] = rule

        return rules

    def __len__(self):
        """Return the number of rules inside this filter"""
        return len(self.rules)

    def __getitems__(self, name: str):
        """Return a specific rule"""
        return self.rules[name]

    def __call__(
        self,
        mols: Sequence[Union[str, dm.Mol]],
        n_jobs: Optional[int] = -1,
        progress: bool = False,
        progress_leave: bool = False,
        scheduler: str = "auto",
        keep_props: bool = False,
        fail_if_invalid: bool = True,
    ) -> pd.DataFrame:
        """Compute the rules for a list of molecules

        Args:
            mols: list of input molecule object.
            n_jobs: number of jobs to run in parallel.
            progress: whether to show progress or not.
            progress_leave: whether to leave the progress bar or not.
            scheduler: which scheduler to use. If "auto", will use "processes" if `len(mols) > 500` else "threads".
            keep_props: whether to keep the properties columns computed by the rules.
            fail_if_invalid: whether to fail if a rule fails or not.

        Returns:
            df: Dataframe where each row is a molecule and each column is a the outcomes of applying self.rules[column].
        """

        if scheduler == "auto":
            if len(mols) > 500:
                scheduler = "processes"
            else:
                scheduler = "threads"

        def _rule_fn(mol: Union[str, dm.Mol]):
            # Convert to mol object if needed
            if isinstance(mol, str):
                mol = dm.to_mol(mol)

            if mol is None:
                if fail_if_invalid:
                    raise ValueError("Molecule is None")
                else:
                    return pd.Series()

            datum = pd.Series()
            datum["mol"] = mol

            # Precompute properties
            props = _compute_batch_props(mol)

            if keep_props:
                datum = pd.concat([datum, pd.Series(props)])

            # Filter with the rules
            datum["pass_all"] = True
            datum["pass_any"] = False
            for rule_name, rule_fn in self.rules.items():
                # Do the computation
                datum[rule_name] = rule_fn(mol, **props)

                datum["pass_all"] &= datum[rule_name]
                datum["pass_any"] |= datum[rule_name]

            return datum

        results = dm.parallelized(
            _rule_fn,
            mols,
            progress=progress,
            n_jobs=n_jobs,
            scheduler=scheduler,
            tqdm_kwargs=dict(
                desc="Filter by rules",
                leave=progress_leave,
            ),
        )
        results = pd.DataFrame(results)

        return results

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def list_available_rules(*query: str):
        """List all the available rules and they properties"""
        df = pd.read_csv(loader.get_data_path("medchem_rule_list.csv"))
        if len(query) > 0:
            query_str = "|".join(query)
            df = df[df["description"].str.contains(query_str)]
        return df

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def list_available_rules_names(*query: str):
        """List only the names of the available rules"""
        df = RuleFilters.list_available_rules(*query)
        return df["name"].tolist()



from ._version import __version__

from typing import TYPE_CHECKING

import os

import importlib


# The below lazy import logic is coming from openff-toolkit:
# https://github.com/openforcefield/openff-toolkit/blob/b52879569a0344878c40248ceb3bd0f90348076a/openff/toolkit/__init__.py#L44

# Dictionary of objects to lazily import; maps the object's name to its module path
_lazy_imports_obj = {
    # version
    "__version__": "medchem._version",
}

# Dictionary of modules to lazily import; maps the modules's name to its path
_lazy_imports_mod = {
    "utils": "medchem.utils",
    "groups": "medchem.groups",
    "catalogs": "medchem.catalogs",
    "constraints": "medchem.constraints",
    "complexity": "medchem.complexity",
    "rules": "medchem.rules",
    "structural": "medchem.structural",
    "functional": "medchem.functional",
    "query": "medchem.query",
}


def __getattr__(name):
    """Lazily import objects from _lazy_imports_obj or _lazy_imports_mod

    Note that this method is only called by Python if the name cannot be found
    in the current module."""
    obj_mod = _lazy_imports_obj.get(name)
    if obj_mod is not None:
        mod = importlib.import_module(obj_mod)
        return mod.__dict__[name]

    lazy_mod = _lazy_imports_mod.get(name)
    if lazy_mod is not None:
        return importlib.import_module(lazy_mod)

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    """Add _lazy_imports_obj and _lazy_imports_mod to dir(<module>)"""
    keys = (*globals().keys(), *_lazy_imports_obj.keys(), *_lazy_imports_mod.keys())
    return sorted(keys)


if TYPE_CHECKING or os.environ.get("MEDCHEM_DISABLE_LAZY_LOADING", "0") == "1":
    # These types are imported lazily at runtime, but we need to tell type
    # checkers what they are.

    from . import utils
    from . import groups
    from . import catalogs
    from . import constraints
    from . import complexity
    from . import rules
    from . import structural
    from . import functional
    from . import query



from typing import Dict
from typing import Callable
from typing import List

from loguru import logger

import datamol as dm

from rdkit.Chem.rdchem import SubstructMatchParameters
from rdkit.Chem.rdmolfiles import MolFragmentToSmiles


class Constraints:
    """Complex query system for performing substructure matches with additional constraints

    !!! example

        ```python
        mol1 = dm.to_mol("CN(C)C(=O)c1cncc(C)c1")
        mol2 = dm.to_mol("c1ccc(cc1)-c1cccnc1")
        core = dm.from_smarts("c1cncc([*:1])c1")
        [atom.SetProp("query", "my_constraints") for atom in core.GetAtoms() if atom.GetAtomMapNum() == 1]
        constraint_fns = dict(my_constraints=lambda x: dm.descriptors.n_aromatic_atoms(x) > 0)
        constraint = Constraints(core, constraint_fns)
        matches = [constraint(mol1), constraint(mol2)] # False, True
        ```
    """

    def __init__(self, core: dm.Mol, constraint_fns: Dict[str, Callable], prop_name: str = "query"):
        """Initialize the constraint matcher

        Args:
            core: the scaffold/query molecule to match against. Needs to be a molecule
            constraint_fns: a dictionary of constraint functions that defines the required constraints after the substructure match
            prop_name: the property name to use in the match at each atom defined by the core
                for further matches against the constraints functions
        """
        self.core = core
        self.prop_name = prop_name
        self.constraint_fns = constraint_fns
        self.atom_to_query = self._initialize()

    def _initialize(self):
        """Initialize the constraint matcher"""
        atom_to_query = dict()
        for a in self.core.GetAtoms():
            if a.HasProp(self.prop_name) and str(a.GetProp(self.prop_name)) in self.constraint_fns:
                atom_to_query[a.GetIdx()] = a.GetProp(self.prop_name)
        return atom_to_query

    @staticmethod
    def validate(mol: dm.Mol, constraints: List["Constraints"]):
        """Validate a list of constraint object against a molecule

        Args:
            mol: the molecule object
            constraints: list of Contraints object to validate against the molecule
        """
        mol = dm.to_mol(mol)

        if mol is None:
            raise ValueError("Input molecule is None")

        for constraint in constraints:
            if not isinstance(constraint, Constraints):
                raise ValueError("Input constraint should be an instance of Constraints")

            if not constraint(mol):
                return False

        return True

    def get_matches(self, mol: dm.Mol, multiple: bool = True):
        """Get matches that respect the constraints in the molecules

        Args:
            mol: input molecule
            multiple: if True, return all the matches, if False, return the first match
        """
        if not isinstance(mol, dm.Mol):
            mol = dm.to_mol(mol)

        params = SubstructMatchParameters()
        params.setExtraFinalCheck(self._check_final)
        matches = mol.GetSubstructMatches(self.core, params)

        if multiple:
            return matches

        return matches[0]

    def has_match(self, mol: dm.Mol):
        """Check if input molecule respect the constraints

        Args:
            mol: input molecule
        """
        return len(self.get_matches(mol)) > 0

    def __call__(self, mol: dm.Mol):
        """Check if input molecule respect the constraints

        Args:
            mol: input molecule
        """
        return self.has_match(mol)

    def _check_final(self, mol: dm.Mol, vect: List[int]):
        """
        Perform a breadth first search over current matches.

        !!! note
            This function is not designed to be used directly but this setup is required to make things work
            Always use `validate` or `match


        Args:
            mol: input molecule
            vect: list of atom indexes to check

        """
        seen = [0] * mol.GetNumAtoms()
        for idx in vect:
            seen[idx] = 1

        # loop over the atoms we care about:
        for idx, qfn in self.atom_to_query.items():
            midx = vect[idx]
            atom = mol.GetAtomWithIdx(midx)
            # now do a breadth-first search from that atom, checking
            # all of its neighbors that aren't in the substructure
            # query:
            stack = [atom]
            connected_atoms = set()
            while stack:
                atom = stack.pop(0)
                connected_atoms.add(atom.GetIdx())
                for neigh in atom.GetNeighbors():
                    neigh_idx = neigh.GetIdx()
                    if not seen[neigh_idx] and neigh_idx not in connected_atoms:
                        stack.append(neigh)
            is_ok = False
            try:
                submol = dm.to_mol(MolFragmentToSmiles(mol, atomsToUse=connected_atoms))
                is_ok = self.constraint_fns[qfn](submol)
            except Exception as e:
                logger.error(e)
                # we can't raise here

            if not is_ok:
                return False
        return True



from typing import Optional
from typing import Union

import functools

import datamol as dm
import numpy as np
import pandas as pd

from rdkit.Chem import GraphDescriptors

from medchem.utils.loader import get_data_path

from ._calc import WhitlockCT
from ._calc import BaroneCT
from ._calc import SMCM
from ._calc import TWC


class ComplexityFilter:
    """Complexity-based molecular filters which compare some molecular complexity metrics to
    precomputed thresholds on large datasets of commercially available compounds.

    The default threshold have been calculated using all commercially available compounds of the zinc-15 dataset.

    !!! example "Usage"
        ```python
        from medchem.complexity import ComplexityFilter
        complexity_filter = ComplexityFilter(limit="99", complexity_metric="bertz")
        complexity_filter("CC(=O)Nc1ccc(cc1)O") # True
        ```
    """

    COMPLEXITY_FNS = {
        "bertz": GraphDescriptors.BertzCT,  # bertz complexity index
        "sas": dm.descriptors.sas,  # sas score
        "qed": dm.descriptors.qed,  # qed score
        "clogp": dm.descriptors.clogp,  # clogp score
        "whitlock": WhitlockCT,  # whitlock complexity index
        "barone": BaroneCT,  # barone complexity index
        "smcm": SMCM,  # synthetic and molecular complexity
        "twc": TWC,  # Total walk count complexity
    }

    def __init__(
        self,
        limit: str = "99",
        complexity_metric: str = "bertz",
        threshold_stats_file: Optional[str] = "zinc_15_available",
    ):
        """
        Default complexity limit is set on at least 1 exceeding metric on the 999th permille level

        Args:
            limit: The complexity percentile outlier limit to be used (should be a string expressed as an integer)
            complexity_metric: The complexity filter name to be used.
                Use `ComplexityFilter.list_default_available_filters` to list default filters.
                The following complexity metrics are supported by default
                * "bertz": bertz complexity index
                * "sas": synthetic accessibility score  (`zinc_15_available` only)
                * "qed": qed score  (`zinc_15_available` only)
                * "clogp": clogp for how greasy a molecule is compared to other in the same mw range  (`zinc_15_available` only)
                * "whitlock": whitlock complexity index
                * "barone": barone complexity index
                * "smcm": synthetic and molecular complexity
                * "twc":  total walk count complexity  (`zinc_15_available` only)
            threshold_stats_file: The path to or type the threshold file to be used.
                The default available threshold stats files are
                * "zinc_12"
                * "zinc_15_available"

        """
        self.threshold_df = self.load_threshold_stats_file(threshold_stats_file)

        if limit not in self.threshold_df.percentile.unique():
            raise ValueError(f"Invalid value {limit}")

        available_filters = set(self.list_default_available_filters()).intersection(self.threshold_df.columns)

        if complexity_metric not in available_filters:
            raise ValueError(f"Invalid value {complexity_metric}. Should be one of {available_filters}")

        self.limit_index = limit
        self.complexity_metric = complexity_metric
        cur_df = self.threshold_df[self.threshold_df.percentile == self.limit_index]
        self.filter_selection_df = cur_df[[self.complexity_metric, "mw_bins"]].sort_values(
            "mw_bins",
            ascending=True,
        )

    @classmethod
    def list_default_available_filters(cls):
        """
        Return a list of unique filter names
        """
        return list(ComplexityFilter.COMPLEXITY_FNS.keys())

    @classmethod
    @functools.lru_cache()
    def list_default_percentile(cls, threshold_stats_file: Optional[str] = None):
        """
        Return the default percentile list for the threshold file
        """
        df = cls.load_threshold_stats_file(threshold_stats_file)
        return df["percentile"].unique().tolist()

    @classmethod
    def load_threshold_stats_file(cls, path: Optional[str] = None):
        """
        Load threshold file to compute the percentille depending on the MW for each complexity_metric
        Args:
            path: path to the threshold file
        """
        if path is None or path == "zinc_12":
            path = "zinc_12_stats.csv"
            path = get_data_path(path, module="medchem.data.complexity")

        elif path == "zinc_15_available":
            path = "zinc_15_available_stats.csv"
            path = get_data_path(path, module="medchem.data.complexity")

        data = pd.read_csv(path)
        return data

    def __call__(self, mol: Union[dm.Mol, str]):
        """
        Check whether the input structure is too complex given this instance of the complexity filter
        Return False is the molecule is too complex, else True

        Args:
            mol: input molecule
        """

        if isinstance(mol, str):
            mol = dm.to_mol(mol)

        mw = dm.descriptors.mw(mol)
        ind = np.digitize(mw, self.filter_selection_df["mw_bins"], right=True)

        fn = ComplexityFilter.COMPLEXITY_FNS[self.complexity_metric]
        threshold = self.filter_selection_df[self.complexity_metric].values[ind]

        with dm.without_rdkit_log():
            pred = fn(mol)

        return np.isnan(pred) or pred < threshold



from medchem.utils.graph import score_symmetry


def test_score_symmetry():
    test_mols = [
        # high symmetry ~0.857
        ("O=C(O)c1cc(-n2ccnc2)cc(-n2ccnc2)c1", False),
        # fully symmetric
        ("OC(O)c1cc(-n2ccnc2)cc(-n2ccnc2)c1", True),
        # symmetry should be 1, modif of above
        ("CC(C)(C)[C@@H]1COC(C2(C3=N[C@H](C(C)(C)C)CO3)Cc3ccccc3C2)=N1", True),
        ("c1ccc2oc(-c3ccc(-c4nc5ccccc5o4)s3)nc2c1", True),
        # symmetrical star graph
        ("CC(C)(C)", True),
        ("NC(C)(N)C", True),
        # non symmetrical
        ("NC(C)(C)C", False),
        # modification of the above. not symmetrical
        ("CCCC1=CNC(CC)=C1C", False),
        # disconnection does not prevent symmetric detection
        ("Cl.[O-]C(=O)CC([O-])=O.CCCC1=CNC=C1CCC", True),
        # regular salt, should be symmetrical
        ("Cl.CC([O-])=O.CCCC1=CNC=C1CCC", True),
        # not a salt, should not be symmetrical
        ("[O-]C(=O)C1=CC=CC=C1.CCCC1=CNC=C1CCC", False),
    ]

    mols, expected_vals = zip(*test_mols)

    scores = [score_symmetry(x, exclude_self_mapped_edged=False) for x in mols]
    fully_symmetrical = [x == 1 for x in scores]

    assert list(expected_vals) == fully_symmetrical

    # if we don't exclude self_mapped edges, then we will have non symmetrical here
    expected_mol2 = score_symmetry(mols[1], exclude_self_mapped_edged=True)
    assert expected_mol2 < 1
    assert expected_mol2 > 0.8



from typing import List

import itertools
import datamol as dm

from rdkit.Chem.rdmolops import ReplaceCore
from rdkit.Chem.rdmolops import GetMolFrags
from rdkit.Chem.Scaffolds import MurckoScaffold

from loguru import logger

from medchem.utils.smarts import SMARTSUtils

_DESCRIPTOR_LIST = [
    "mw",
    "fsp3",
    "n_lipinski_hba",
    "n_lipinski_hbd",
    "n_rings",
    "n_hetero_atoms",
    "n_heavy_atoms",
    "n_rotatable_bonds",
    "n_radical_electrons",
    "n_NHOH",
    "n_NO",
    "tpsa",
    "qed",
    "clogp",
    "sas",
    "formal_charge",
    "refractivity",
    "n_aliphatic_carbocycles",
    "n_aliphatic_heterocycles",
    "n_aliphatic_rings",
    "n_aromatic_carbocycles",
    "n_aromatic_heterocycles",
    "n_aromatic_rings",
    "n_saturated_carbocycles",
    "n_saturated_heterocycles",
    "n_saturated_rings",
    "n_charged_atoms",
    "n_stereo_center",
    "n_aromatic_atoms",
    "n_rigid_bonds",
    "n_aromatic_atoms_proportion",
]


def in_range(
    x: float,
    min_val: float = -float("inf"),
    max_val: float = float("inf"),
):
    """Check if a value is in a range

    Args:
        x: value to check
        min_val: minimum value
        max_val: maximum value
    """
    return min_val <= x <= max_val


def n_heavy_metals(mol: dm.Mol, allowed_metals: List[str] = ["Li", "Be", "K", "Na", "Ca", "Mg"]):
    """Count the number of heavy metals in a molecule

    Metal atoms are defined using the M notation in marvinjs. It's quicker to exclude atoms than to list all metals

    Args:
        mol: input molecule
        allowed_metals: list of metals not counted as heavy metals. Default is ["Li", "Be", "K", "Na", "Ca", "Mg"]
    """
    non_metals = [
        2,
        5,
        6,
        7,
        8,
        9,
        10,
        14,
        15,
        16,
        17,
        18,
        33,
        34,
        35,
        36,
        52,
        53,
        54,
        85,
        86,
    ]

    heavy_metals = [
        x.GetAtomicNum()
        for x in mol.GetAtoms()
        if x.GetAtomicNum() not in non_metals and x.GetSymbol() not in allowed_metals
    ]
    return len(heavy_metals)


def has_spider_chains(mol: dm.Mol, min_appendage: int = 2, min_appendage_len: int = 4):
    """Check whether a molecule has multiple appendage-like structures

    Args:
        mol: input molecule
        min_appendage: minimum number of appendages (>=)
        min_appendage_len: minimum length (number of atoms in straight line) of a appendage (>=)
    """
    # we first need to remove all rings systems (so scaffold from the molecules)
    # then we can count the both condition for spider/appendage like look
    side_chains = [mol]
    appendage_query = SMARTSUtils.aliphatic_chain(min_size=min_appendage_len)
    scaffold = MurckoScaffold.GetScaffoldForMol(mol)
    try:
        scaffold = dm.sanitize_mol(scaffold)
        side_chains = ReplaceCore(mol, scaffold, labelByIndex=False)
        if side_chains is not None:
            side_chains = list(GetMolFrags(side_chains, asMols=True))
            side_chains = [dm.to_smiles(x) for x in side_chains]
            side_chains = [SMARTSUtils.standardize_attachment(x, "[1*]") for x in side_chains]
            side_chains = [dm.to_mol(x) for x in side_chains]
            appendage_query = "[1*]-,=" + appendage_query
        else:
            side_chains = [mol]
    except Exception as e:
        logger.error(e)

    assert all([isinstance(x, dm.Mol) for x in side_chains])

    # extract side chains from the scaffold
    appendage_query = dm.from_smarts(appendage_query)

    if appendage_query is None:
        raise ValueError("Invalid appendage query")

    matches = []
    for x in side_chains:
        if x is None or isinstance(x, str):
            raise ValueError("None molecule in side chains")

        matches.append(x.HasSubstructMatch(appendage_query))

    return sum(matches) >= min_appendage


def n_fused_aromatic_rings(mol: dm.Mol, require_all_aromatic: bool = True, pairwise: bool = False):
    """Count the number of fused aromatic rings in a molecule

    !!! warning
        There is no such thing as a spiroaoaromatic ring in this implementation

    Args:
        mol: input molecule
        require_all_aromatic: whether to require all simple rings in the fused system to be aromatic
        pairwise: whether to compute the number of fused aromatic rings pairwise.
            meaning phenanthrene and anthracene would count for 2 fused aromatic rings each
    """

    # EN: might make sense to move this to datamol
    # This code can be spedt up by sacrificing readability, will revisit eventually

    ring_systems = mol.GetRingInfo()

    # we use bond since we are focusing on fused rings
    simple_rings = list(ring_systems.BondRings())
    rings = [set(r) for r in simple_rings]
    ring_map = [set([x]) for x in range(len(rings))]

    go_next = True
    while go_next:
        go_next = False
        for i, j in itertools.combinations(range(len(rings)), 2):
            if rings[i] & rings[j]:
                new_map = set().union(ring_map[i], ring_map[j])
                q = rings[i] | rings[j]
                min_ind, max_ind = min(i, j), max(i, j)
                del rings[max_ind], rings[min_ind]
                del ring_map[max_ind], ring_map[min_ind]
                rings.append(q)
                ring_map.append(new_map)
                go_next = True
                break

    # simple_rings: is the list of simple rings from ring info
    # rings: the list of rings after mergin fused rings
    # ring_map: the mapping between fused rings and the basic rings their contains
    if pairwise:
        # we need to count all pair in any fused rings with more than 2 fused rings
        fused_rings = []
        for _, ring_ids in enumerate(ring_map):
            for ring_1, ring_2 in itertools.combinations(ring_ids, 2):
                if set(simple_rings[ring_1]) & set(simple_rings[ring_2]):
                    fused_ring = set.union(set(simple_rings[ring_1]), (simple_rings[ring_2]))
                    fused_rings.append(fused_ring)
    else:
        fused_rings = [r for i, r in enumerate(rings) if len(ring_map[i]) >= 2]

    n_aromatic_fused_rings = 0
    for i, fused_ring_bonds in enumerate(fused_rings):
        aromatic_system = [mol.GetBondWithIdx(bond_id).GetIsAromatic() for bond_id in fused_ring_bonds]

        if require_all_aromatic:
            n_aromatic_fused_rings += all(aromatic_system)
        else:
            n_aromatic_fused_rings += any(aromatic_system)

    return n_aromatic_fused_rings


def fraction_atom_in_scaff(mol: dm.Mol):
    """Compute the fraction of atoms that belong to any ring system of a molecule
    as defined by its murcko scaffold

    Args:
        mol: input molecule
    """
    n_heavy_atoms = mol.GetNumHeavyAtoms()
    if n_heavy_atoms < 1:
        return 0

    n_heavy_scaffold_atoms = 0
    scaffold = MurckoScaffold.GetScaffoldForMol(mol)
    scaffold = dm.sanitize_mol(scaffold)

    if scaffold is not None:
        n_heavy_scaffold_atoms = scaffold.GetNumHeavyAtoms()

    return n_heavy_scaffold_atoms / n_heavy_atoms


def list_descriptors():
    """List all descriptors available for computation"""
    return _DESCRIPTOR_LIST



import datamol as dm
import medchem as mc


def test_num_atom_filter():
    data = dm.data.freesolv()
    data = data.iloc[:10]
    input_list = data["smiles"].apply(dm.to_mol).tolist()

    passing_idx = [i for i, m in enumerate(input_list) if m.GetNumAtoms() < 20]
    idx = mc.functional.num_atom_filter(input_list, max_atoms=20, return_idx=True)

    assert idx.tolist() == passing_idx


def test_ring_infraction_filter():
    smiles_with_issues = [
        # none
        "Nc1noc2cc(-c3noc(C(F)(F)F)n3)ccc12",
        # hologenicity, atom_list_filter, ring_infraction
        "Nc1noc2cc(-c3noc(n3)C(Cl)(Cl)Cl)c(Cl)c(B3C=C3)c12",
        # num_atom_filter
        "C[C@H](Nc1ncc(-c2noc(C(F)(F)F)n2)cc1Cl)c1ccncc1",
        # macrocycle, num_atom_filter
        "NC1=NOC2=CC(C3=NOC(=N3)C(F)(F)F)=C3CCCCCCCC3=C12",
        # ring infraction, macrocycle at 10, num_atom_filter at 20
        "Nc1noc2c(cc(cc12)C1NCO1)C1CCCCCCCCC1",
    ]

    out = mc.functional.ring_infraction_filter(smiles_with_issues, return_idx=True)
    out = list(out)
    assert [0, 2, 3] == out


def test_macrocycle_filter():
    smiles_with_issues = [
        # none
        "Nc1noc2cc(-c3noc(C(F)(F)F)n3)ccc12",
        # hologenicity, atom_list_filter, ring_infraction
        "Nc1noc2cc(-c3noc(n3)C(Cl)(Cl)Cl)c(Cl)c(B3C=C3)c12",
        # num_atom_filter
        "C[C@H](Nc1ncc(-c2noc(C(F)(F)F)n2)cc1Cl)c1ccncc1",
        # macrocycle, num_atom_filter
        "NC1=NOC2=CC(C3=NOC(=N3)C(F)(F)F)=C3CCCCCCCC3=C12",
        # ring infraction, macrocycle at 10, num_atom_filter at 20
        "Nc1noc2c(cc(cc12)C1NCO1)C1CCCCCCCCC1",
    ]

    out = mc.functional.macrocycle_filter(smiles_with_issues, max_cycle_size=7, return_idx=False)
    out = list(out)
    assert out == [True, True, True, False, False]


def test_atom_list_filter():
    smiles_with_issues = [
        # none
        "Nc1noc2cc(-c3noc(C(F)(F)F)n3)ccc12",
        # hologenicity, atom_list_filter, ring_infraction
        "Nc1noc2cc(-c3noc(n3)C(Cl)(Cl)Cl)c(Cl)c(B3C=C3)c12",
        # num_atom_filter
        "C[C@H](Nc1ncc(-c2noc(C(F)(F)F)n2)cc1Cl)c1ccncc1",
        # macrocycle, num_atom_filter
        "NC1=NOC2=CC(C3=NOC(=N3)C(F)(F)F)=C3CCCCCCCC3=C12",
        # ring infraction, macrocycle at 10, num_atom_filter at 20
        "Nc1noc2c(cc(cc12)C1NCO1)C1CCCCCCCCC1",
    ]

    out = mc.functional.atom_list_filter(smiles_with_issues, unwanted_atom_list=["B"], return_idx=True)
    out = list(out)
    assert [0, 2, 3, 4] == out


def test_halogenicity_filter():
    smiles_with_issues = [
        # none
        "Nc1noc2cc(-c3noc(C(F)(F)F)n3)ccc12",
        # hologenicity, atom_list_filter, ring_infraction
        "Nc1noc2cc(-c3noc(n3)C(Cl)(Cl)Cl)c(Cl)c(B3C=C3)c12",
        # num_atom_filter
        "C[C@H](Nc1ncc(-c2noc(C(F)(F)F)n2)cc1Cl)c1ccncc1",
        # macrocycle, num_atom_filter
        "NC1=NOC2=CC(C3=NOC(=N3)C(F)(F)F)=C3CCCCCCCC3=C12",
        # ring infraction, macrocycle at 10, num_atom_filter at 20
        "Nc1noc2c(cc(cc12)C1NCO1)C1CCCCCCCCC1",
    ]

    out = mc.functional.halogenicity_filter(smiles_with_issues, return_idx=True)
    out = list(out)
    assert [0, 2, 3, 4] == out


def test_n_stereo_center():
    smiles = [
        "CC(C)(O)C(O)[C@H](O)C(O)CO",  # fail, too many undefined stereocenter
        "C1[C@H]([C@@H]([C@H]([C@H](O1)O)O)O)O",  # fail, too many stereocenter
        "C[C@@H]1CC[C@H]([C@@H](C1)O)C(C)C",  # pass
    ]

    out = mc.functional.num_stereo_center_filter(smiles, return_idx=False)
    out = list(out)
    assert [False, False, True] == out


def test_symmetry_filter():
    symmetric_mols = [
        "O=C(O)c1cc(-n2ccnc2)cc(-n2ccnc2)c1",
        "CC(C)(C)[C@@H]1COC(C2(C3=N[C@H](C(C)(C)C)CO3)Cc3ccccc3C2)=N1",
        "c1ccc2oc(-c3ccc(-c4nc5ccccc5o4)s3)nc2c1",
        "Cc1cc(O)c(C(c2ccc(Cl)cc2)c2c(O)cc(C)[nH]c2=O)c(=O)[nH]1",
    ]

    random_mols = [
        "CCn1cc(S(=O)(=O)n2cc(Cl)cn2)cn1",
        "Cc1cc(C)c(S(=O)(=O)Nc2cc(C)ccc2C)c(C)c1",
        "c1ccc(CCC2CCN(CCC3COCCO3)CC2)cc1",
        "CCCC1CCC([C@H]2CC[C@H](C(=O)O)CC2)CC1",
    ]
    symmetric = mc.functional.symmetry_filter(symmetric_mols, symmetry_threshold=0.8)
    assert symmetric.tolist() == [False] * len(symmetric)

    not_symmetric = mc.functional.symmetry_filter(random_mols, symmetry_threshold=0.8)
    assert not_symmetric.tolist() == [True] * len(not_symmetric)



import math

from rdkit.Chem.rdmolops import GetMolFrags
from rdkit.Chem.rdmolops import FindPotentialStereo
from rdkit.Chem import FindMolChiralCenters

import datamol as dm


def WhitlockCT(
    mol: dm.Mol,
    ringval: float = 4,
    unsatval: float = 2,
    heteroval: float = 1,
    chiralval: float = 2,
):
    """
    A chemically intuitive measure for molecular complexity. This complexity measure
    has been described in : H. W. Whitlock, J. Org. Chem., 1998, 63, 7982-7989.
    Benzyls, fenyls, etc. are not treated at all.

    On the zinc 15 commercially available dataset, the range of this score is [0, 172] with a median of 25

    Args:
        mol: The input molecule.
        ringval: The contribution of rings
        unsatval: The contribution of the unsaturated bond.
        heteroval: The contribution of the heteroatom.
        chiralval: The contribution of the chiral center.
    """

    # possibly TODO phenyls and others are considered as protecting group thus could be removed (keyword argument removePhenyl)
    mol = dm.to_mol(mol, add_hs=True)

    if mol is None:
        raise ValueError("Invalid molecule")

    n_bonds = 0
    n_unsat = 0
    n_hetero = 0
    n_chiral = len(FindMolChiralCenters(mol, includeUnassigned=True))
    n_arom = 0

    bonds = mol.GetBonds()
    atoms = mol.GetAtoms()

    for bond in bonds:
        n_bonds += 1
        btype = bond.GetBondType()
        if btype == dm.DOUBLE_BOND:
            n_unsat += 1
            n_bonds += 1
        elif btype == dm.TRIPLE_BOND:
            n_unsat += 2
            n_bonds += 2
        elif btype == dm.AROMATIC_BOND:
            n_arom += 1

    for atom in atoms:
        if atom.GetAtomicNum() != 1 and atom.GetAtomicNum() != 6:
            n_hetero += 1

    n_rings = len(bonds) - len(atoms) + len(GetMolFrags(mol))
    complexity = ringval * n_rings + unsatval * n_unsat + heteroval * n_hetero + chiralval * n_chiral

    return complexity


def BaroneCT(mol: dm.Mol, chiral: bool = False):
    """
    Compute the Barone complexity measure for a molecule as described in:
    [R. Barone and M. Chanon, J. Chem. Inf. Comput. Sci., 2001, 41 (2), pp 269–272](https://pubs.acs.org/doi/abs/10.1021/ci000145p)

    On zinc 15 commercially available dataset, the range of this score is [30, 4266] with a median of 538

    Args:
        mol: The input molecule.
        chiral: Whether to include chirality in the calculation.
    """
    mol = dm.to_mol(mol)
    cmpx = 0
    rinfo = mol.GetRingInfo()
    for aring in rinfo.AtomRings():
        cmpx += len(aring) * 6
    for atom in mol.GetAtoms():
        deg = atom.GetExplicitValence()
        if deg == 1:
            cmpx += 3
        elif deg == 2:
            cmpx += 6
        elif deg == 3:
            cmpx += 12
        else:
            cmpx += 24
        if atom.GetAtomicNum() != 6:
            cmpx += 6
        else:
            cmpx += 3
    if chiral:
        cmpx += 20 * len(FindMolChiralCenters(mol, includeUnassigned=True))
    return cmpx


_SMCM_ENs = {
    5: 0.851,
    6: 1,
    7: 1.149,
    8: 1.297,
    9: 1.446,
    15: 1.086,
    16: 1.235,
    17: 1.384,
    35: 1.244,
    53: 1.103,
}

_SMCM_BNs = {
    dm.SINGLE_BOND: {
        6: {
            6: 1.0,
            7: 0.857,
            8: 0.75,
            9: 0.667,
            15: 0.4,
            16: 0.375,
            17: 0.353,
            35: 0.171,
            53: 0.113,
        },
        7: {7: 0.735, 8: 0.643},
        8: {16: 0.281},
    },
    dm.DOUBLE_BOND: {
        6: {
            6: 0.5,
            7: 0.429,
            8: 0.375,
            16: 0.188,
        },
        7: {7: 0.367, 8: 0.321},
        8: {16: 0.141},
    },
    dm.TRIPLE_BOND: {6: {6: 0.333, 7: 0.286}},
    dm.AROMATIC_BOND: {
        6: {6: 0.667, 7: 0.571, 16: 0.25},
        7: {7: 0.49, 8: 0.423},
    },
}
"""
Table of bond
atom    atom    single    double    triple    aromatic
C        C      1.000     0.500     0.333     0.667
C        N      0.857     0.429     0.286     0.571
C        O      0.750     0.375
C        F      0.667
C        P      0.400
C        S      0.375     0.188               0.250
C        Cl     0.353
C        Br     0.171
C        I      0.113
N        N      0.735     0.367               0.490
N        O      0.643     0.321               0.423
O        S      0.281     0.141
"""


def _SMCM_GetBondScore(btype, a1, a2):
    ats = sorted([a1, a2])
    s = 0
    try:
        s = _SMCM_BNs[btype][ats[0]][ats[1]]
    except KeyError:
        pass
    return s


# SMARTS patterns for SMCM metric which lower structure complexity
_SMCM_SMARTSs = [
    dm.from_smarts(patt)
    for patt in (
        "[$([C!D1]-@C(=O)-@[N!a])]",  # peptides, cyclic
        "[$([C!D1]-!@C(=O)-!@[N!a])]",  # peptides, acyclic
        "[$([#6]=,:[*R!C!c]=,:[#6R]~@[*R!C!c])]",  # C attached to Hetero in ring
        "[$([A!C!H](-!@[#6])-!@[#6])]",  # hetero attached to 2 carbons not in ring
        "[$([#7]=[C!$(*a)]-!@[N!H0])]",  # amidine, guanidine
        "[$([#8]=[#6H0]-!@[N!H2])!$(NC[!#6])]",  # nonterminal amide
        "[#8]=[#6](-!@[NH1,NH0])-!@[N!H2]",  # nonterminal urea
        "[$(O=[CD3]([#6])[#6])!$([!#6]=CC=O)]",  # ketone, not diketo
        "[$([#16X2v2]([#6])[#6])]",  # thio ether
        "[$([#8X2H0][#6]=[#8D1])!$(O~C(~O)~O)]",  # carboxyl ester, not carbonate
        "[$([#8X2v2]([#6])[#6])]",  # ether oxygen
        "C1:c-@C-@N-@C-@C-@O-@C-@c:c-1",  # SMARTS from ICCB's Diversity-Oriented
        "[N,O,C]C(=O)C1=C[C@H](*)C[C@H](O)O1",  # Synthesis approach19
        "C[C@H]1O[C@H]~3O[C@H]~C~2~C~C[C@H]([C@@H]~1)[C@@H]~23",  #
        "C[C@@H]1O[C@@H]~2O[C@H][C@@H][C@@H]~3~C~C~C~1[C@@H]~23",  #
        "C[C@H]2C[C@]14~C~C~C~C[C@H]1Oc3cccc(CN2)c34",  #
        "[#6][C@H]1C[C@@H]([#6])O[C@@H](-a)O1",  #
        "C2=CC[C@@H]1C(=O)~*~*~C(=O)[C@@H]1[C@@H]2-a",  #
        "C2=CC[C@@H]1C(=O)~*~C(=O)[C@@H]1[C@@H]2-a",  #
        "a-[CH,CH2;R0;0*]",  # from ref 17
        "[R;0*]-[CH2R0,NHR0,OR0;0*]-[R]",  #
        "*-[CD3H,ND2;R0;0*](-a)-a",  #
        "[a]-&!@[a]",  #
        "[NR;0*]-[CD3R0;0*](=O)-[R]",  #
        "[NR;0*]-[CD2R0;0*]-[R]",  #
        "[NR;0*]-[CD2R0;0*]-[CD2,CD3,OD2,ND2,ND3,aD2,aD3]",  #
        "a-[NHR0]-[CR0;0*](=O)-[OR0,NR0,0*]",  #
        "[CR,NR]=[CR]-&!@[a]",  #
        "[$([#6](-!@[#7])-!@[#7!H0])]",  # NHCN not in ring
        "[$([#6](@[#7])@[#7!H0])]",  # NHCN in ring
        "[$([A!C!H](-@[#6])-@[#6])]",  # hetero attached to 2 carbons in ring
        "[$([#6](=[#8])-[#6]=[!#6])]",  # diketo, keto-Ryl
        "[$([#16X3v4,#16X4v6]([#6])[A])]",  # hypervalent sulfur
        "[$(C(~O)(~O)~O)]",  # carbonate
    )
]


def SMCM(mol: dm.Mol):
    """
    Compute synthetic and molecular complexity as described in:
    [TK Allu, TI Oprea, J. Chem. Inf. Model. 2005, 45(5), pp. 1237-1243](https://pubs.acs.org/doi/10.1021/ci0501387)

    On zinc 15 commercially available dataset, the range of this score is [1.93, 192.00] with a median of 42.23

    Args:
        mol: the input molecule
    """
    mol = dm.to_mol(mol)
    ats = mol.GetAtoms()
    bonds = mol.GetBonds()
    a_score = 0
    c = 0
    for a in ats:
        if a.GetAtomicNum() == 6:
            c += 1
        if a.GetAtomicNum() in _SMCM_ENs:
            a_score += _SMCM_ENs[a.GetAtomicNum()]
        else:
            a_score += 1
            # TODO log it or print this in the same way as RDKit
            print("Unsupported atom #{}".format(a.GetAtomicNum()))

    b_score = 0
    for b in bonds:
        b_score += _SMCM_GetBondScore(
            b.GetBondType(),
            b.GetBeginAtom().GetAtomicNum(),
            b.GetEndAtom().GetAtomicNum(),
        )

    chiral_a = len(FindPotentialStereo(mol)) * 2
    mot_score = sum([0] + [len(mol.GetSubstructMatches(patt)) for patt in _SMCM_SMARTSs])
    score = a_score + b_score + chiral_a - mot_score

    return round(score, 3)


def _AWC(k: int, atom: int, table: list, neighbors: list):
    """Compute walk count for atom"""
    if atom not in table[k]:
        table[k][atom] = 0
        for i in neighbors[atom]:
            table[k][atom] += _AWC(k - 1, i, table, neighbors)
    return table[k][atom]


def TWC(mol: dm.Mol, log10: bool = True):
    """Compute total walk count in a molecules as proxy for complexity. This score is described in:
    [Gerta Rucker and Christoph Rucker, J. Chem. Inf. Comput. Sci. 1993, 33, 683-695](https://pubs.acs.org/doi/pdf/10.1021/ci00015a005)

    The total walk count is defined as: $twc = \\frac{1}{2} \sum_{k=1}^{n-1} \sum_{i}^{Natoms} \\text{awc}(k,i)$
    where $\\text{awc}(k,i)$ is the number of walk of length `k` starting at atom `i`.

    On zinc 15 commercially available dataset, the range of this score is [1.20, 39.08] with a median of 10.65

    Args:
        mol: the input molecule
        log10: whether to return the log10 of the values
    """
    mol = dm.to_mol(mol)
    twc = 0
    table = []
    neighbors = []
    atoms = mol.GetAtoms()
    for k in range(len(atoms)):
        table.append({})
    for atom in atoms:
        neighbors.append([i.GetIdx() for i in atom.GetNeighbors()])
        table[0][atom.GetIdx()] = len(neighbors[atom.GetIdx()])

    for k in range(len(atoms) - 1):
        for i in atoms:
            twc += _AWC(k, i.GetIdx(), table, neighbors)
    twc /= 2
    if log10:
        try:
            return math.log10(twc)
        except ValueError:
            return float("nan")
    return twc



from lark import Transformer, v_args


class QueryParser(Transformer):
    """
    Query parser for the custom query language for molecule. This parses the input language, build a parseable and evaluable representation.
    The trick for lazy evaluation is to define custom guard with '`fn(*)`' around expression that needs to be evaluated.

    Note that you **SHOULD NOT HAVE TO INTERACT WITH THIS CLASS DIRECTLY**.

    Example:
        ```python
        import medchem
        import lark
        QUERY_GRAMMAR = medchem.utils.loader.get_grammar(as_string=True)
        QUERY_PARSER = Lark(QUERY_GRAMMAR, parser="lalr", transformer=QueryParser())
        # see how the string needs to be "quoted". This builds on the json quote requirements to avoid dealing with unwanted outcomes
        example = \"""(HASPROP("tpsa" > 120 ) | HASSUBSTRUCTURE("c1ccccc1")) AND NOT HASALERT("pains") OR HASSUBSTRUCTURE("[OH]", max)\"""
        t = QUERY_PARSER.parse(example)
        print(t)
        ((((`fn(getprop, prop='tpsa')` > 120.0) or `fn(hassubstructure, query='c1ccccc1', operator='None', limit=None, is_smarts=None)`) and not `fn(hasalert, alert='pains')`) or `fn(hassubstructure, query='[OH]', operator='max', limit=None, is_smarts=None)`)
        ```
    """

    @v_args(inline=True)
    def not_bool_factor(self, *args):
        """Define representation of a negation"""
        return " ".join([str(x) for x in args])

    @v_args(inline=True)
    def bool_expr(self, bool_term, *others):
        """Define how boolean expressions should be parsed"""
        or_clauses = " ".join(f"{other}" for other in others)
        return f"({bool_term} {or_clauses})"

    @v_args(inline=True)
    def bool_term(self, bool_factor, *others):
        """Define how boolean terms should be parsed"""
        and_clauses = " ".join(f"{other}" for other in others)
        return f"({bool_factor} {and_clauses})"

    @v_args(inline=True)
    def hassubstructure(self, value, is_smarts, operator, limit):
        """Format the substructure node in the query

        !!! note
            The parser does not enforce any validity on the argument and
            the underlying function is supposed to handle it.

        """
        return f"`fn(hassubstructure, query='{value}', is_smarts={is_smarts}, operator={operator}, limit={limit})`"

    @v_args(inline=True)
    def hassuperstructure(self, value):
        """Format the superstructure node in the query

        !!! note
            The parser does not enforce any validity on the argument and
            the underlying function is supposed to handle it.

        """
        return f"`fn(hassuperstructure, query='{value}')`"

    @v_args(inline=True)
    def hasalert(self, value):
        """Format the hasalert node in the query

        !!! note
            The parser does not enforce any validity on the argument and
            the underlying function is supposed to handle it.

        """
        return f"`fn(hasalert, alert='{value}')`"

    @v_args(inline=True)
    def hasgroup(self, value):
        """Format the hasgroup node in the query

        !!! note
            The parser does not enforce any validity on the argument and
            the underlying function is supposed to handle it.

        """
        return f"`fn(hasgroup, group='{value}')`"

    @v_args(inline=True)
    def matchrule(self, value):
        """Format the matchrule node in the query

        !!! note
            The parser does not enforce any validity on the argument and
            the underlying function is supposed to handle it.

        """
        return f"`fn(matchrule, rule='{value}')`"

    @v_args(inline=True)
    def hasprop(self, value, comparator, limit):
        """Format the hasprop node in the query

        !!! note
            The parser does not enforce any validity on the argument and
            the underlying function is supposed to handle it.

        """
        return f"(`fn(getprop, prop='{value}')` {comparator} {limit})"

    @v_args(inline=True)
    def like(self, value, comparator, limit):
        """Format the like node in the query

        !!! note
            The parser does not enforce any validity on the argument and
            the underlying function is supposed to handle it.

        """
        return f"(`fn(similarity, query='{value}')` {comparator} {limit})"

    # non function call
    @v_args(inline=True)
    def operator(self, value):
        if value is None:
            return "None"
        return f"'{value}'"

    @v_args(inline=True)
    def comparator(self, value):
        if value == "=":
            # equality fix
            value = "=="
        return f"{value}"

    def is_smarts(self, value):
        return bool(value)

    def TRUE(self, _):
        return True

    def FALSE(self, _):
        return False

    def MIN(self, _):
        return "min"

    def MAX(self, _):
        return "max"

    def NOT_OP(self, _):
        return "not"

    def AND_OP(self, _):
        return "and"

    def OR_OP(self, _):
        return "or"

    def ESCAPED_STRING(self, value):
        return value[1:-1]

    def INT(self, i):
        return int(i)

    def SIGNED_NUMBER(self, f):
        return float(f)



from typing import List
from typing import Optional
from typing import Union

import os
import functools
import pandas as pd
import datamol as dm

from rdkit.Chem import GetMolFrags
from rdkit.Chem.rdmolops import ReplaceCore
from rdkit.Chem.rdmolops import AdjustQueryParameters
from rdkit.Chem.rdmolops import AdjustQueryProperties
from rdkit.Chem.rdmolops import AdjustQueryWhichFlags

from medchem.catalogs import catalog_from_smarts
from medchem.utils.loader import get_data_path


def list_default_chemical_groups(hierarchy: bool = False) -> list:
    """List all the available chemical groups.

    !!! note
        chemical groups defines how a collection of patterns are organized.
        They **do not** correspond to individual pattern name.

    Args:
        hierarchy: whether to return the full hierarchy or the group name only

    Returns:
        List of chemical groups
    """
    data = pd.read_csv(get_data_path("chemical_groups.csv"))
    if hierarchy:
        return list(data.hierarchy.unique())
    return list(data.group.unique())


def list_functional_group_names(unique: bool = True) -> list:
    """
    List common functional group names

    Args:
        unique: whether to return only unique names

    Returns:
        List of functional group names
    """
    data = pd.read_csv(get_data_path("chemical_groups.csv"))
    data = data[data.hierarchy.str.contains("functional_groups")]
    return list(data.name.unique())


@functools.lru_cache(maxsize=None)
def get_functional_group_map() -> dict:
    """
    Map functional groups to their corresponding SMARTS string.

    Returns:
        List of functional group names
    """
    data = pd.read_csv(get_data_path("chemical_groups.csv"))
    data = data[data.hierarchy.str.contains("functional_groups")]
    # EN: any group that is not unique should be dropped
    # this is because of the `basic_groups` hierarchy that is loosely defined
    data = data.drop_duplicates(subset=["name"], keep=False)
    data = data.sort_values("name", ascending=True)
    return dict(zip(data["name"], data["smarts"]))


class ChemicalGroup:
    """Build a library of chemical groups using a list of structures parsed from a file

    The default library of structure has been curated from https://github.com/Sulstice/global-chem and additional open source data.

    !!! note
        For new chemical groups, please minimally provide the 'smiles'/'smarts', 'name' and "group" and optional 'hierarchy' columns

    !!! warning
        The SMILES and SMARTS used in the default list of substructures do not result in the same matches.
        Unless specified otherwise, the SMILES will be used in the matching done by this class,
        whereas due to RDKit's limitation, the SMARTS will be used in the matching done by the generated catalog.

    """

    def __init__(
        self,
        groups: Optional[Union[str, List[str]]] = None,
        n_jobs: Optional[int] = None,
        groups_db: Optional[Union[os.PathLike, str]] = None,
    ):
        """Build a chemical group library

        Args:
            groups: List of groups to use. Defaults to None where all functional groups are used
            n_jobs: Optional number of jobs to run in parallel for internally building the data. Defaults to None.
            groups_db: Path to a file containing the dump of the chemical groups. Default is internal dataset
        """

        if isinstance(groups, str):
            groups = [groups]
        if groups is None:
            groups = []
        self.groups = groups
        self.n_jobs = n_jobs or 0
        if groups_db is None:
            groups_db = get_data_path("chemical_groups.csv")
        self.data = pd.read_csv(groups_db)
        if "hierarchy" not in self.data.columns:
            self.data["hierarchy"] = self.data["group"]
        if self.groups:
            self.data = self.data[self.data.hierarchy.str.contains("|".join(self.groups))]
        # EN: fill smiles and smarts with empty string when they are missing
        # this prevent error when applying dm.to_mol, dm.from_smarts and building the FilterCatalog
        self.data["smiles"] = self.data["smiles"].fillna("")
        self.data["smarts"] = self.data["smarts"].fillna("")
        self._initialize_data()

    def filter(self, names: List[str], fuzzy: bool = False):
        """Filter the group to restrict to only the name in input

        Args:
            names: list of names to use for filters
            fuzzy: whether to use exact or fuzzy matching
        """
        if names is None or len(names) == 0:
            return self

        if fuzzy:
            names = [f".*{name}.*" for name in names]
            self.data = self.data[self.data["name"].str.match(r"|".join(names))]
        else:
            self.data = self.data[self.data["name"].isin(names)]

        return self

    def _initialize_data(self):
        """Initialize the data by precomputing some features"""
        with dm.without_rdkit_log():
            self.data["mol_smarts"] = dm.parallelized(
                dm.from_smarts,
                self.data["smarts"].values,
                n_jobs=self.n_jobs,
                progress=False,
            )
            self.data["mol"] = dm.parallelized(
                dm.to_mol,
                self.data["smiles"].values,
                n_jobs=self.n_jobs,
                progress=False,
            )

    def __len__(self):
        return len(self.data)

    @property
    def name(self):
        """Get the Name of the chemical groups in this instance"""
        return self.data.name.tolist()

    @property
    def smiles(self):
        """Get the SMILES of the chemical groups in this instance"""
        return self.data.smiles.tolist()

    @property
    def mols(self):
        """Get the Molecule object of the SMILES for the chemical groups in this instance"""
        return self.data.mol.tolist()

    @property
    def smarts(self):
        """Get the SMARTS of the chemical groups in this instance"""
        return self.data.smarts.tolist()

    @property
    def mol_smarts(self):
        """Get the SMARTS of the chemical groups in this instance"""
        return self.data.mol_smarts.tolist()

    @property
    def mol_adjusted(self):
        """Get the Molecules object of the SMILES, adjusted for stricter match for the chemical groups in this instance"""
        # EN: beware of pickling issues with AdjustQueryParameters
        query_adjust = AdjustQueryParameters()
        query_adjust.adjustRingChain = True
        query_adjust.adjustHeavyDegree = True
        query_adjust.adjustDegree = True
        query_adjust.adjustDegreeFlags = (
            AdjustQueryWhichFlags.ADJUST_IGNORERINGS | AdjustQueryWhichFlags.ADJUST_IGNORECHAINS
        )
        mols = self.data.mol.apply(lambda x: AdjustQueryProperties(x, query_adjust))
        return mols.tolist()

    @property
    def dataframe(self):
        """Get the dataframe of the chemical groups"""
        return self.data

    # @functools.lru_cache(maxsize=32)
    def get_catalog(self, exact_match: bool = True):
        """Build an rdkit catalog from the current chemical group data

        Args:
            exact_match: whether to adjust the queries for a more stringent match
        """

        if exact_match:
            return catalog_from_smarts(
                self.mol_adjusted,
                self.name,
                entry_as_inds=False,
            )
        return catalog_from_smarts(
            self.mol_smarts,
            self.name,
            entry_as_inds=False,
        )

    def list_groups(self):
        """List all the chemical groups available"""
        return list(self.data.group.unique())

    def list_hierarchy_groups(self):
        """List all the hierarchy in chemical groups available.
        To get the full hierarchy on each path, split by the `.` character.
        """
        return list(self.data.hierarchy.unique())

    def get_matches(
        self,
        mol: Union[dm.Mol, str],
        use_smiles: bool = True,
        exact_match: bool = False,
        terminal_only: bool = False,
    ):
        """Get all the functional groups in this instance that matches the input molecule

        Args:
            mol: input molecule
            use_smiles: whether to use the smiles representation of the catalog or the smarts
            exact_match: whether to use exact matching by adjusting the query
            terminal_only: ensure whether the matches to the functional group are terminal,
                meaning that any subgraph matching should not be in the middle of the molecules.
        """
        if isinstance(mol, str):
            mol = dm.to_mol(mol)

        if mol is None:
            return None

        # EN: beware of pickling issues with AdjustQueryParameters
        query_adjust = AdjustQueryParameters()
        query_adjust.adjustRingChain = True
        query_adjust.adjustDegree = True

        def matcher(query):
            if exact_match:
                query = AdjustQueryProperties(query, query_adjust)
            matches = mol.GetSubstructMatches(query, uniquify=True) or None
            if matches is not None and terminal_only:
                side_chains = [ReplaceCore(mol, query, match) for match in matches]
                side_chains = [GetMolFrags(x, asMols=True) for x in side_chains]
                matches = [x for x, y in zip(matches, side_chains) if len(y) == 1]
            return matches or None

        if use_smiles:
            matches = dm.parallelized(matcher, self.data.mol.values, n_jobs=self.n_jobs, progress=False)
        else:
            matches = dm.parallelized(
                matcher, self.data.mol_smarts.values, n_jobs=self.n_jobs, progress=False
            )
        out = self.data[["name", "smiles", "smarts", "group"]].copy()
        out["matches"] = matches
        out = out.dropna(subset=["matches"])
        return out

    def has_match(self, mol: Union[dm.Mol, str], exact_match: bool = False, terminal_only: bool = False):
        """Check whether the input molecule has any functional group in this instance

        Args:
            mol: input molecule
            exact_match: whether to use exact matching by adjusting the query
            terminal_only: ensure the matches to the functional group are terminal
        """
        matches = self.get_matches(mol, exact_match=exact_match, terminal_only=terminal_only)
        return matches is not None and len(matches) > 0



import pytest

import medchem as mc
import datamol as dm

from medchem.structural.lilly_demerits import LillyDemeritsFilters


def test_common_alerts():
    alerts = mc.structural.CommonAlertsFilters()

    data = dm.data.freesolv()
    data = data.iloc[:50]

    data["mol"] = data["smiles"].apply(dm.to_mol)

    results = alerts(
        mols=data["mol"].tolist(),
        n_jobs=-1,
        scheduler="processes",
        keep_details=True,
    )

    assert results["pass_filter"].sum() == 44
    assert results["reasons"].unique().tolist() == [
        None,
        "halogen_heteroatom;sulfonyl_halide",
        "primary_halide_sulfate",
        "non_ring_CH2O_acetal;phosphorus_sulfur_bond",
        "aldehyde",
        "gte_10_carbon_sb_chain;gte_8_CF2_or_CH2",
    ]

    assert set(results.columns.tolist()) == {"mol", "pass_filter", "status", "reasons", "details"}


def test_common_alerts_invalid():
    alerts = mc.structural.CommonAlertsFilters()

    results = alerts(mols=[None, "CC9888", "CCCCO"])

    assert results["mol"].isnull().sum() == 2
    assert results.shape == (3, 4)
    assert results["reasons"].tolist() == ["invalid", "invalid", None]


def test_common_alerts_list():
    ll = mc.structural.CommonAlertsFilters.list_default_available_alerts()
    assert ll.columns.tolist() == ["rule_set_name", "smarts", "catalog_description", "rule_set", "source"]


def test_nibr():
    nibr_filters = mc.structural.NIBRFilters()

    data = dm.data.solubility()
    data = data.iloc[:50]

    results = nibr_filters(
        mols=data["mol"].tolist(),
        n_jobs=-1,
        scheduler="processes",
        keep_details=True,
    )

    assert results["pass_filter"].sum() == 49
    assert set(results.columns.tolist()) == {
        "mol",
        "reasons",
        "severity",
        "status",
        "n_covalent_motif",
        "special_mol",
        "pass_filter",
        "details",
    }


def test_nibr_invalid():
    nibr_filters = mc.structural.NIBRFilters()

    results = nibr_filters(mols=[None, "CC9888", "CCCCO"])

    assert results["mol"].isnull().sum() == 2
    assert results.shape == (3, 7)
    assert results["reasons"].tolist() == ["invalid", "invalid", None]
    assert set(results.columns.tolist()) == {
        "mol",
        "reasons",
        "severity",
        "status",
        "n_covalent_motif",
        "special_mol",
        "pass_filter",
    }


def test_lilly_demerits():
    dfilters = LillyDemeritsFilters()

    data = dm.data.solubility()
    data = data.sample(50, random_state=20)

    results = dfilters(mols=data["mol"].tolist())

    assert results["pass_filter"].sum() == 30
    assert set(results.columns.tolist()) == {
        "smiles",
        "reasons",
        "step",
        "demerit_score",
        "status",
        "pass_filter",
        "mol",
    }


def test_lilly_demerits_config():
    test_config = {
        "output": "test",
        "min_atoms": 7,
        "soft_max_atoms": 30,
        "hard_max_atoms": 50,
        "smarts": [],
        "nodemerit": False,
        "dthresh": 160,
        "odm": [],
        "okiso": False,
        "noapdm": False,
    }

    dfilters = LillyDemeritsFilters(**test_config)

    data = dm.data.solubility()
    data = data.sample(50, random_state=20)

    results = dfilters(mols=data["mol"].tolist())

    assert results["pass_filter"].sum() == 30
    assert set(results.columns.tolist()) == {
        "smiles",
        "reasons",
        "step",
        "demerit_score",
        "status",
        "pass_filter",
        "mol",
    }


def test_demerits_invalid():
    dfilters = LillyDemeritsFilters()

    with pytest.raises(ValueError):
        dfilters(mols=[None, "CC9888", "CCCCO"])



from typing import Optional
from typing import Union
from typing import Any
from typing import Sequence

import os
import uuid
import re
import shutil
import tempfile
import importlib.resources as importlib_resources

import pandas as pd
import datamol as dm
import numpy as np


from ._lilly import find_lilly_binaries
from ._lilly import parse_output
from ._lilly import run_cmd


BIN2PATH = find_lilly_binaries()


class LillyDemeritsFilters:
    """Lilly MedChem Rules published in:

    [Robert F. Bruns and Ian A. Watson, Rules for Identifying Potentially Reactive or Promiscuous Compounds,
    Journal of Medicinal Chemistry 2012 55 (22), 9763-9772](https://pubs.acs.org/doi/10.1021/jm301008n)


    !!! abstract "Description"
        This is a set of 275 rules, developed over an 18-year period, used to identify compounds that may interfere with biological assays,
        allowing their removal from screening sets. Reasons for rejection include reactivity (e.g., acyl halides),
        interference with assay measurements (fluorescence, absorbance, quenching), activities that damage proteins (oxidizers, detergents),
        instability (e.g., latent aldehydes), and lack of druggability (e.g., compounds lacking both oxygen and nitrogen).


    """

    def __init__(
        self,
        mc_first_pass_options: Optional[str] = None,
        iwd_options: Optional[str] = None,
        stop_after_step: int = 3,
        **run_options: Any,
    ):
        """
        Constructor for the Lilly MedChem Rules

        Args:
            mc_first_pass_options: Initial options to pass to mc_first_pass
            iwd_options: Initial options to pass to iwdemerit
            stop_after_step: Where to stop in the pipeline. Don't change this if you don't know.
            run_options: Additional option to run the pipeline
        """
        self.mc_first_pass_options = mc_first_pass_options
        self.iwd_options = iwd_options
        self.stop_after_step = stop_after_step
        self.run_options = run_options

    def __call__(
        self,
        mols: Sequence[Union[str, dm.Mol]],
        n_jobs: Optional[int] = -1,
        batch_size: int = 5_000,
        progress: bool = False,
        progress_leave: bool = False,
        scheduler: str = "threads",
    ):
        if scheduler == "processes":
            raise ValueError("Demerits filtering does not support processes or loky mode yet.")

        n_splits = max(1, int(np.ceil(len(mols) / batch_size)))

        if n_splits > 1:
            mols_batch_list = np.array_split(mols, n_splits)
            # EN: cannot run this code in processes or loky mode
            results = dm.parallelized(
                self._score,
                mols_batch_list,
                n_jobs=n_jobs,
                progress=progress,
                scheduler=scheduler,
                tqdm_kwargs=dict(
                    desc="Demerits filtering",
                    leave=progress_leave,
                ),
            )
            return pd.concat(results, ignore_index=True)  # type: ignore

        else:
            return self._score(mols)

    def _score(self, mols: Sequence[Union[str, dm.Mol]]):
        """Run lilly medchem scorer on input smile list:

        Args:
            mols: list of smiles
        """

        mc_first_pass_options = self.mc_first_pass_options
        iwd_options = self.iwd_options
        stop_after_step = self.stop_after_step
        run_options = self.run_options

        if mc_first_pass_options is None:
            mc_first_pass_options = ""

        if iwd_options is None:
            iwd_options = ""

        extra_iwdemerit_options = ""
        demerit_cutoff = run_options.get("dthresh", None)
        soft_upper_atom = run_options.get("soft_max_atoms", 30)
        hard_upper_atom = run_options.get("hard_max_atoms", 50)
        max_size_rings = run_options.get("max_size_rings")
        min_num_rings = run_options.get("min_num_rings")
        max_num_rings = run_options.get("max_num_rings")
        max_size_chain = run_options.get("max_size_chain")
        atom_count = run_options.get("min_atoms", 1)
        allow_non_int_atoms = run_options.get("allow_non_interesting", False)

        ring_bond_ratio = run_options.get("ring_bond_ratio", -1)
        okiso = run_options.get("okiso", False)
        if not okiso:
            mc_first_pass_options += " -I 0 "
        if min_num_rings:
            mc_first_pass_options += f"-r {min_num_rings} "
        if max_num_rings:
            mc_first_pass_options += f"-R {max_num_rings} "
        if max_size_rings:
            mc_first_pass_options += f"-Z {max_size_rings} "
            extra_iwdemerit_options += f" -Z {max_size_rings} "
        if max_size_chain:
            extra_iwdemerit_options += f" -z {max_size_chain} "

        if allow_non_int_atoms:
            mc_first_pass_options += "-k "
        mc_first_pass_options += " -A I -A ipp"

        query_files = ["reject1", "reject2", "demerits"]
        for i, query_file in enumerate(query_files):
            query_files[i] = str(importlib_resources.files("medchem.data.queries").joinpath(query_file))

        # output file dir
        files_to_be_deleted = []
        run_id = str(uuid.uuid4())[:8]
        bad_file_dir = tempfile.mkdtemp(suffix=f"_lilly_{run_id}")
        files_to_be_deleted.append(bad_file_dir)
        bad_file_0 = os.path.join(bad_file_dir, "bad0")
        bad_file_1 = os.path.join(bad_file_dir, "bad1")
        bad_file_2 = os.path.join(bad_file_dir, "bad2")
        bad_file_3 = os.path.join(bad_file_dir, "bad3")
        mc_pass_out = os.path.join(bad_file_dir, "mc_pass.smi")
        tsub_out_1 = os.path.join(bad_file_dir, "tsub1.smi")
        tsub_out_2 = os.path.join(bad_file_dir, "tsub2.smi")
        iwd_out = os.path.join(bad_file_dir, "iwd.smi")

        # optional_queries
        optional_queries = ""
        optional_queries += " -q ".join(run_options.get("rej_queries", []))
        optional_queries += " -s ".join(run_options.get("smarts", []))

        # extra iwdemerit options
        nodemerit = run_options.get("nodemerit", False)

        if nodemerit or soft_upper_atom is None:
            soft_upper_atom = hard_upper_atom - 1

        if hard_upper_atom < soft_upper_atom:
            hard_upper_atom = soft_upper_atom + 1

        if demerit_cutoff:
            extra_iwdemerit_options += f" -f {demerit_cutoff} "

        if nodemerit:
            extra_iwdemerit_options += " -r "

        if iwd_options:
            extra_iwdemerit_options += " " + iwd_options

        charge_assigner_path = str(
            importlib_resources.files("medchem.data.charge_assigner").joinpath("queries")
        )
        extra_iwdemerit_options += " -N F:" + charge_assigner_path

        odm = run_options.get("odm", [])

        if odm:
            odm_patterns = [re.compile(odm_p, re.I) for odm_p in odm]
            with open(query_files[2]) as QRY_IN:
                current_queries = [qry.strip() for qry in QRY_IN]
                current_queries = [
                    qry for qry in current_queries if not any(odm_p.search(qry) for odm_p in odm_patterns)
                ]

            with tempfile.NamedTemporaryFile(mode="w+t", suffix=".qry", delete=False) as tmp_file:
                tmp_file.write("\n".join(current_queries))
                query_files[2] = tmp_file.name
                files_to_be_deleted.append(tmp_file.name)

        smiles_file = None
        with tempfile.NamedTemporaryFile(
            mode="w+t",
            suffix=f"_lilly_{run_id}.smi",
            delete=False,
        ) as smiles_tmp_files:
            # Convert the input mols to a list of SMILES
            smiles_list = []
            for mol in mols:
                if isinstance(mol, str):
                    smiles = mol
                else:
                    smiles = dm.to_smiles(mol)

                # Sanity check
                if smiles is None or dm.to_mol(smiles) is None:
                    raise ValueError(f"Invalid SMILES: {smiles}. Demerits does not support invalid mol yet.")

                smiles_list.append(smiles)

            smiles_tmp_files.write(
                "\n".join([f"{sm.strip().split()[0]}\t{i}" for i, sm in enumerate(smiles_list)])
            )
            smiles_file = smiles_tmp_files.name
            files_to_be_deleted.append(smiles_file)

        cmd = [BIN2PATH["mc_first_pass"]]
        if ring_bond_ratio >= 0:
            cmd.extend(["-b", str(ring_bond_ratio)])

        if mc_first_pass_options:
            cmd.extend(mc_first_pass_options.split())

        cmd.extend(["-c", str(atom_count), "-C", str(hard_upper_atom)])
        cmd.extend("-E autocreate -o smi -V -g all -g ltltr -i ICTE".split())
        cmd.extend(["-L", bad_file_0, "-K", "TP1"])
        cmd.extend(["-a", "-u", "-S", mc_pass_out, smiles_file])

        out = []
        out.append(run_cmd(cmd))

        if stop_after_step >= 1:
            cmd = []
            cmd.extend((BIN2PATH["tsubstructure"] + " -E autocreate -b -u -i smi -o smi -A D ").split())
            cmd.extend(("-m " + bad_file_1 + " -m QDT").split())
            cmd.extend(
                (f"-n {tsub_out_1} -q F:" + query_files[0] + optional_queries + f" {mc_pass_out}").split()
            )
            out.append(run_cmd(cmd))

        if stop_after_step >= 2:
            cmd = []
            cmd.extend((BIN2PATH["tsubstructure"] + " -A D -E autocreate -b -u -i smi -o smi ").split())
            cmd.extend(("-m " + bad_file_2 + " -m QDT").split())
            cmd.extend((f"-n {tsub_out_2} -q F:" + query_files[1] + f" {tsub_out_1}").split())
            out.append(run_cmd(cmd))

        if stop_after_step >= 3:
            cmd = []
            cmd.extend(
                (
                    BIN2PATH["iwdemerit"]
                    + " -u -k -x -t "
                    + extra_iwdemerit_options
                    + " -E autocreate -A D -i smi -o smi -q F:"
                    + query_files[2]
                ).split()
            )
            cmd.extend(f"-R {bad_file_3}".split())
            cmd.extend(
                f"-G {iwd_out} -c smax={soft_upper_atom} -c hmax={hard_upper_atom} {tsub_out_2}".split()
            )
            out.append(run_cmd(cmd))

        data_list = []
        for i, bad_file in enumerate([bad_file_0, bad_file_1, bad_file_2]):
            with open(bad_file + ".smi") as IN:
                df_bad = parse_output(IN)
                df_bad["step"] = i + 1
                df_bad["rejected"] = True
                data_list.append(df_bad)

        demerit_extractor = re.compile(r"'([A-Za-z0-9_\s]+)'")
        i = 0
        for dt_file, rej in [(bad_file_3 + ".smi", True), (iwd_out, False)]:
            parseable = []
            demerit_scores = []
            with open(dt_file) as IN:
                for row in IN:
                    in_string, *demerit_string = row.split(" : ")
                    in_string = in_string.strip()
                    demerit_score = 0
                    if not demerit_string:
                        in_string += ' ""'
                    else:
                        demerit_string = str(demerit_string[0]).strip()
                        m = re.match(r"D\(([0-9]+)\)", demerit_string)
                        if m is not None:
                            demerit_score = int(m.group(1))
                        tmp = demerit_extractor.findall(row)
                        demerit_string = ",".join([":".join(reversed(line.split())) for line in tmp])
                        in_string += " " + demerit_string
                    parseable.append(in_string)
                    demerit_scores.append(demerit_score)

            df = parse_output(parseable)
            df["rejected"] = rej
            df["step"] = i + 1
            df["demerit_score"] = demerit_scores
            data_list.append(df)
            i += 1
        results = pd.concat(data_list).sort_values("ID").reset_index(drop=True)

        # Postprocessing
        results["status"] = results["rejected"].apply(lambda x: "Exclude" if x else "Ok")
        results.loc[((results.demerit_score > 0) & (~results.rejected)), "status"] = "Flag"

        results["status"] = results["status"].str.lower()
        results["pass_filter"] = ~results["rejected"]
        results = results.drop(columns=["ID", "rejected"])
        results["mol"] = mols

        # clean
        for to_del in files_to_be_deleted:
            if os.path.isfile(to_del):
                os.remove(to_del)
            else:
                shutil.rmtree(to_del)

        return results



import io
import re
import shutil
import subprocess

from loguru import logger

import pandas as pd


def find_lilly_binaries():
    """Find the required binaries to run the Lilly medchem filters"""
    binaries_list = ["mc_first_pass", "tsubstructure", "iwdemerit"]
    binary_paths = {}
    for binary_name in binaries_list:
        binary_path = shutil.which(binary_name)

        if binary_path is None:
            raise ImportError(
                "The Lilly binaries required to use the `medchem.structural.demerits` module seems to be missing. "
                "Install with `mamba install lilly-medchem-rules`."
            )

        binary_paths[binary_name] = binary_path

    return binary_paths


def rreplace(input_str, old, rep, occurrence):
    """Replace last occurence of 'old' in 'input_str' by 'rep'"""
    tmp = input_str.rsplit(old, occurrence)
    return rep.join(tmp)


def parse_output(rowlist):
    """Parse content of `rowlist` to dataframe"""
    content = "\n".join(
        [
            rreplace(
                re.sub(r"\s+\([0-9]*\s+(matches\sto\s)", ' "', line.strip(), 1),
                "')",
                "'\"",
                1,
            ).strip("'")
            for line in rowlist
        ]
    )
    flux = io.StringIO(content)
    df = pd.read_csv(flux, sep=" ", doublequote=True, names=["smiles", "ID", "reasons"])
    df["ID"] = pd.to_numeric(df["ID"])
    df["reasons"] = df["reasons"].apply(lambda x: x.strip("'") if x and isinstance(x, str) else x)
    return df


def run_cmd(cmd, shell=False):
    """Run a command"""
    res = subprocess.run(cmd, capture_output=True, shell=shell, check=False)
    if res.returncode != 0:
        logger.error("".join(res.stderr.decode("utf-8")))
        logger.error(" ".join(cmd))
        res.check_returncode()
    return res



from typing import Callable
from typing import Union

import types
import functools


def get_function_name(fn: Union[types.FunctionType, Callable, object]) -> str:
    """Get the name of a function"""

    if isinstance(fn, functools.partial):
        fn = fn.func
    if isinstance(fn, types.FunctionType):
        return fn.__name__
    elif isinstance(fn, object):
        return type(fn).__name__

    raise ValueError(f"Cannot find name of input function: {fn}")



from ._api import alert_filter
from ._api import nibr_filter
from ._api import catalog_filter
from ._api import chemical_group_filter
from ._api import rules_filter
from ._api import complexity_filter
from ._api import bredt_filter
from ._api import molecular_graph_filter
from ._api import lilly_demerit_filter
from ._api import protecting_groups_filter

from ._generic import macrocycle_filter
from ._generic import atom_list_filter
from ._generic import ring_infraction_filter
from ._generic import num_atom_filter
from ._generic import num_stereo_center_filter
from ._generic import halogenicity_filter
from ._generic import symmetry_filter



from typing import Optional

import datamol as dm

import re

ATTACHMENT_POINT_TOKEN = re.escape("*")
# The following allows discovering and extracting valid dummy atoms in smiles/smarts.
ATTACHMENT_POINTS = [
    # parse any * not preceeded by "[" or ":"" and not followed by "]" or ":" as attachment
    r"(?<![:\[]){0}(?![:\]])".format(ATTACHMENT_POINT_TOKEN),
    # fix some edge cases of the above for attachment in the "()" environment or in SMARTS notation
    r"(?<![\[(\[\w)]){0}".format(ATTACHMENT_POINT_TOKEN),
    # parse attachment in optionally mapped isotope smarts atoms. eg. [1*:3]
    r"\[(\d*){}:?(\d*)\]".format(ATTACHMENT_POINT_TOKEN),
]


class SMARTSUtils:
    """Collections of utils to build complex SMARTS query more efficiently for non experienced user"""

    @classmethod
    def standardize_attachment(cls, smiles: str, attach_tokens: str = "[*:1]"):
        """Standardize an attachment point in a smiles

        Args:
            smiles: SMILES string
            attach_tokens: Attachment point token to use as standard token
        """
        if not smiles:
            return smiles

        for attach_regex in ATTACHMENT_POINTS:
            smiles = re.sub(attach_regex, attach_tokens, smiles)

        return smiles

    @classmethod
    def ortho(
        cls,
        smarts_str1: str,
        smarts_str2: str,
        aromatic_only: bool = False,
    ) -> str:
        """
        Returns a recursive smarts string connecting the two input smarts in `ortho` of each other.
        Connection points needs to be through single or double bonds.

        Args:
            smarts_str1: first smarts pattern defining the first functional group
            smarts_str2: second smarts pattern defining the second functional group
            aromatic_only: whether the ring needs to be aromatic or not

        Example:
            to build a smarts for a methyl group in ortho to an oxygen (e.g: 'C1CC(C)C(O)CC1')
            >>> SMARTSUtils.ortho('[#6;!R]', '[#8]')

        Returns:
            smarts: smarts pattern connecting the two input smarts in `ortho` of each other
        """
        if aromatic_only:
            return f"[$({smarts_str1})]-,=!:aa-,=!:[$({smarts_str2})]"
        return f"[$({smarts_str1})]-,=!:[r][r]-,=!:[$({smarts_str2})]"

    @classmethod
    def meta(
        cls,
        smarts_str1: str,
        smarts_str2: str,
        aromatic_only: bool = False,
    ) -> str:
        """
        Returns a recursive smarts string connecting the two input smarts in `meta` of each other.
        Connection points needs to be through single or double bonds.

        Args:
            smarts_str1: first smarts pattern defining the first functional group
            smarts_str2: second smarts pattern defining the second functional group
            aromatic_only: whether the ring needs to be aromatic or not

        Example:
            to build a smarts for a methyl group in meta to an oxygen (e.g: 'c1c(C)cc(O)cc1')
            >>> SMARTSUtils.meta('[#6;!R]', '[#8]')

        Returns:
            smarts: smarts pattern connecting the two input smarts in `meta` of each other
        """
        if aromatic_only:
            return f"[$({smarts_str1})]-,=!:aaa-,=!:[$({smarts_str2})]"

        return f"[$({smarts_str1})]-,=!:[r][r][r]-,=!:[$({smarts_str2})]"

    @classmethod
    def para(
        cls,
        smarts_str1: str,
        smarts_str2: str,
        aromatic_only: bool = False,
    ) -> str:
        """
        Returns a recursive smarts string connecting the two input smarts in `para` of each other.
        Connection points needs to be through single or double bonds.

        Args:
            smarts_str1: first smarts pattern defining the first functional group
            smarts_str2: second smarts pattern defining the second functional group
            aromatic_only: whether the ring needs to be aromatic or not

        Example:
            to build a smarts for a methyl group in para to an oxygen (e.g: 'c1(C)ccc(O)cc1')
            >>> SMARTSUtils.para('[#6;!R]', '[#8]')

        Returns:
            smarts: smarts pattern connecting the two input smarts in `para` of each other
        """
        if aromatic_only:
            return f"[$({smarts_str1})]-,=!:aaaa-,=!:[$({smarts_str2})]"
        return f"[$({smarts_str1})]-,=!:[r][r][r][r]-,=!:[$({smarts_str2})]"

    @classmethod
    def aliphatic_chain(
        cls,
        min_size: int = 6,
        unbranched: bool = False,
        unsaturated_bondtype: Optional[str] = None,
        allow_hetero_atoms: bool = True,
    ) -> str:
        """
        Returns a query that can match a long aliphatic chain

        Args:
            min_size: minimum size of the long chain
            unbranched: whether the chain should be unbranched
            unsaturated_bondtype: additional unsaturated bond type to use for the query. By default, Any bond type (~) is used.
                Single bonds ARE always allowed and bondtype cannot be aromatic
            allow_hetero_atoms: whether the chain can contain hetero atoms

        Example:
            to build a query for a long aliphatic chain of a least 5 atoms (e.g: 'CCC(C)CCC')
            >>> SMARTSUtils.aliphatic_chain(min_size=5)

        Returns:
            smarts: smarts pattern matching a long aliphatic chain
        """
        if unsaturated_bondtype in [dm.AROMATIC_BOND, ":"]:
            raise ValueError("Cannot use aromatic bonds for aliphatic chain")

        include_hets = "" if allow_hetero_atoms else "#6;"
        base_query = "[{}AR0]".format(include_hets)
        bond = "~"
        if unbranched:
            base_query = "[{}R0;D2,D1]".format(include_hets)  # allow terminal too
        if unsaturated_bondtype in [dm.SINGLE_BOND, "-"]:
            bond = ""
        elif unsaturated_bondtype in [dm.DOUBLE_BOND, "="]:
            bond = "-,="
        elif unsaturated_bondtype in [dm.TRIPLE_BOND, "#"]:
            bond = "-,#"

        query = bond.join([base_query] * min_size)
        return query

    @classmethod
    def atom_in_env(cls, *smarts_strs: str, include_atoms: bool = False, union: bool = False) -> str:
        """
        Returns a recursive/group smarts to find an atom that fits in the environments defined by all the input smarts

        Args:
            smarts_strs: list of input patterns defining the environment the atom must fit in. The first atom of each pattern
                should be the atom we want to match to, unless include_atoms is set to True, then [*:99] will be added at the start of each pattern
            include_atoms: whether to include an additional first atom that needs to be in the required environment or not
            union: whether to use the union of the environments or the intersection

        Example:
            you can use this function to construct a complex query if you are not sure about how to write the smarts
            for example, to find a carbon atom that is both in a ring or size 6, bonded to an ethoxy and have a Fluorine in meta

            ```python
            SMARTSUtils.atom_in_env("[#6;r6][OD2][C&D1]", "[c]aa[F]", union=False) # there are alternative way to write this
            ```

        Returns:
            smarts: smarts pattern matching the group/environment
        """

        _smarts_strs = list(smarts_strs)

        if include_atoms:
            _smarts_strs = [f"[*:99]{sm}" for sm in _smarts_strs]

        _smarts_strs = [f"$({sm})" for sm in _smarts_strs]

        if union:
            query = ",".join(_smarts_strs)
        else:
            query = ";".join(_smarts_strs)

        if query:
            query = f"[{query}]"
        return query



from typing import List
from typing import Union
from typing import Optional
from typing import Sequence

import functools

from loguru import logger

import pandas as pd
import numpy as np
import datamol as dm

from rdkit.Chem import rdfiltercatalog  # type: ignore

from medchem.utils.loader import get_data_path


def list_named_catalogs():
    """
    List all available named catalogs. This list will not report chemical groups !

    !!! tip
        For a list of chemical groups that can be queried using `NamedCatalog.chemical_groups`,
        use `medchem.groups.list_default_chemical_groups`
    """
    return [
        x
        for x in NamedCatalogs.__dict__.keys()
        if (not x.startswith("_") and x not in ["alerts", "chemical_groups"])
    ]


def merge_catalogs(*catalogs):
    """Merge several catalogs into a single one

    Returns:
        catalog (FilterCatalog): merged catalog
    """
    if len(catalogs) == 1:
        return catalogs[0]
    params = rdfiltercatalog.FilterCatalogParams()
    missing_catalogs = []
    for catlg in catalogs:
        if isinstance(catlg, rdfiltercatalog.FilterCatalogParams.FilterCatalogs):
            params.AddCatalog(catlg)
        else:
            missing_catalogs.append(catlg)
    parameterized_catalogs = rdfiltercatalog.FilterCatalog(params)
    for catlg in missing_catalogs:
        for entry_nums in range(catlg.GetNumEntries()):
            entry = catlg.GetEntryWithIdx(entry_nums)
            parameterized_catalogs.AddEntry(entry)
    return parameterized_catalogs


def catalog_from_smarts(
    smarts: Union[Sequence[str], np.ndarray, pd.Series],
    labels: Optional[Union[Sequence[str], np.ndarray, pd.Series]] = None,
    mincounts: Optional[Union[Sequence[int], np.ndarray, pd.Series]] = None,
    maxcounts: Optional[Union[Sequence[int], np.ndarray, pd.Series]] = None,
    entry_as_inds: bool = False,
):
    """Load catalog from a list of smarts

    Args:
        smarts: list of input smarts to add to the catalog
        labels: list of label for each smarts
        mincounts: minimum count before a match is recognized
        maxcounts: maximum count for a match to be valid
        entry_as_inds: whether to use index for entry id or the label

    Returns:
        catalog (FilterCatalog): filter catalog built from the input smarts
    """

    with dm.without_rdkit_log():
        catalog = rdfiltercatalog.FilterCatalog()
        if labels is None:
            labels = smarts
        if mincounts is None:
            mincounts = [1] * len(smarts)

        for i, (sm, lb, count) in enumerate(zip(smarts, labels, mincounts)):
            if maxcounts is None:
                fil = rdfiltercatalog.SmartsMatcher(lb, sm, count)
            else:
                fil = rdfiltercatalog.SmartsMatcher(lb, sm, count, maxcounts[i])
            entry_name = str(lb)
            if entry_as_inds:
                entry_name = str(i)

            if fil.IsValid():
                catalog.AddEntry(rdfiltercatalog.FilterCatalogEntry(entry_name, fil))
            else:
                logger.warning(f"It seem like the SMARTS {sm} is invalid")
    return catalog


class NamedCatalogs:
    """
    Holder for substructure matching catalogs. This class provides several popular and custom catalog
    that can be used for substructure matching.

    All the catalogs are cached using `functools.lru_cache` to avoid reloading them every time.

    !!! note
        A filter catalog is a collection of substructures and molecular patterns (SMARTS) used to flag molecules with **(un)desirable structural properties**.
    """

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def tox(
        pains_a: bool = True,
        pains_b: bool = True,
        pains_c: bool = False,
        brenk: bool = True,
        nih: bool = False,
        zinc: bool = False,
    ):
        """Common toxicity and interference catalog

        Args:
            pains_a: whether to include PAINS filters (assay A)
            pains_b: whether to include PAINS filters (assay B)
            pains_c: whether to include PAINS filters (assay C)
            brenk: whether to include BRENK filters, also known as Dundee filters
            nih: whether to include NIH filters
            zinc: whether to include ZINC filters
        """
        catalogs = []
        if pains_a:
            catalogs.append(rdfiltercatalog.FilterCatalogParams.FilterCatalogs.PAINS_A)
        if pains_b:
            catalogs.append(rdfiltercatalog.FilterCatalogParams.FilterCatalogs.PAINS_B)
        if pains_c:
            catalogs.append(rdfiltercatalog.FilterCatalogParams.FilterCatalogs.PAINS_C)
        if brenk:
            catalogs.append(rdfiltercatalog.FilterCatalogParams.FilterCatalogs.BRENK)
        if nih:
            catalogs.append(rdfiltercatalog.FilterCatalogParams.FilterCatalogs.NIH)
        if zinc:
            catalogs.append(rdfiltercatalog.FilterCatalogParams.FilterCatalogs.ZINC)
        catalog = merge_catalogs(*catalogs)
        return catalog

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def pains():
        return rdfiltercatalog.FilterCatalog(rdfiltercatalog.FilterCatalogParams.FilterCatalogs.PAINS)

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def pains_a():
        return rdfiltercatalog.FilterCatalog(rdfiltercatalog.FilterCatalogParams.FilterCatalogs.PAINS_A)

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def pains_b():
        return rdfiltercatalog.FilterCatalog(rdfiltercatalog.FilterCatalogParams.FilterCatalogs.PAINS_B)

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def pains_c():
        return rdfiltercatalog.FilterCatalog(rdfiltercatalog.FilterCatalogParams.FilterCatalogs.PAINS_C)

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def nih():
        return rdfiltercatalog.FilterCatalog(rdfiltercatalog.FilterCatalogParams.FilterCatalogs.NIH)

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def zinc():
        return rdfiltercatalog.FilterCatalog(rdfiltercatalog.FilterCatalogParams.FilterCatalogs.ZINC)

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def brenk():
        return rdfiltercatalog.FilterCatalog(rdfiltercatalog.FilterCatalogParams.FilterCatalogs.BRENK)

    @staticmethod
    def alerts(subset: Optional[Union[List[str], str]] = None):
        """Common alerts filter catalogs commonly used in molecule filtering

        Args:
            subset: subset of alert providers to consider

        Returns:
            catalog (FilterCatalog): filter catalog
        """
        rd_filters = pd.read_csv(get_data_path("common_alerts_collection.csv"))
        if subset is not None:
            if isinstance(subset, str):
                subset = [subset]
            subset = [x.lower() for x in subset]
            rd_filters = rd_filters[rd_filters.rule_set_name.str.lower().isin(subset)]
        mincount = np.maximum(rd_filters["mincount"], 1).astype(int)

        labels = rd_filters.apply(
            lambda x: "{0}||{1}_min({2})||{3}".format(
                x["rule_set_name"],
                x["description"],
                x["mincount"],
                x["priority"],
            ),
            axis=1,
        )
        return catalog_from_smarts(
            rd_filters["smarts"],
            labels,
            mincount,
            entry_as_inds=False,
        )

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def dundee():
        return NamedCatalogs.alerts(subset=["Dundee"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def bms():
        return NamedCatalogs.alerts(subset=["BMS"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def glaxo():
        return NamedCatalogs.alerts(subset=["Glaxo"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def schembl():
        return NamedCatalogs.alerts(subset=["SureChEMBL"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def mlsmr():
        return NamedCatalogs.alerts(subset=["MLSMR"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def inpharmatica():
        return NamedCatalogs.alerts(subset=["Inpharmatica"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def lint():
        return NamedCatalogs.alerts(subset=["LINT"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def alarm_nmr():
        return NamedCatalogs.alerts(subset=["Alarm-NMR"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def alphascreen():
        return NamedCatalogs.alerts(subset=["AlphaScreen-Hitters"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def gst_hitters():
        return NamedCatalogs.alerts(subset=["GST-Hitters"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def his_hitters():
        return NamedCatalogs.alerts(subset=["HIS-Hitters"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def luciferase():
        return NamedCatalogs.alerts(subset=["LuciferaseInhibitor"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def dnabinder():
        return NamedCatalogs.alerts(subset=["DNABinder"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def chelator():
        return NamedCatalogs.alerts(subset=["Chelator"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def hitters():
        return NamedCatalogs.alerts(subset=["Frequent-Hitter"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def electrophilic():
        return NamedCatalogs.alerts(subset=["Electrophilic"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def carcinogen(include_non_genotoxic: bool = True):
        catalogs = ["Genotoxic-Carcinogenicity"]
        if include_non_genotoxic:
            catalogs += ["Non-Genotoxic-Carcinogenicity"]
        return NamedCatalogs.alerts(subset=catalogs)

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def ld50_oral():
        return NamedCatalogs.alerts(subset=["LD50-Oral"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def reactive_unstable_toxic():
        return NamedCatalogs.alerts(subset=["Reactive-Unstable-Toxic"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def skin():
        return NamedCatalogs.alerts(subset=["Skin"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def toxicophore():
        return NamedCatalogs.alerts(subset=["Toxicophore"])

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def nibr():
        """Catalog from NIBR

        !!! warning
            This will return all the compounds in the catalog, regardless of their severity (`FLAG`, `EXCLUDE`, `ANNOTATION`).
            You likely don't want to use this for blind prioritization.
        """
        nibr_filters = pd.read_csv(get_data_path("nibr.csv"))
        mincount = np.maximum(nibr_filters["mincount"], 1).astype(int)
        labels = nibr_filters.apply(
            lambda x: "{0}||{1}_min({2})||{3}||{4}||{5}".format(
                x["rule_set_name"],
                x["description"],
                x["mincount"],
                x["severity"],
                x["covalent"],
                x["special_mol"],
            ),
            axis=1,
        )
        return catalog_from_smarts(
            nibr_filters["smarts"],
            labels,
            mincount,
            entry_as_inds=False,
        )

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def bredt():
        """Bredt Filter Rules: a catalog for filtering unstable molecules, ideal for molecules generated by
        deep learning models or chemical space enumeration.
        See example of usage by [surge](https://github.com/StructureGenerator/SURGE/blob/main/doc/surge1_0.pdf)
        """
        bredt_df = pd.read_csv(get_data_path("bredt.csv"))
        return catalog_from_smarts(
            bredt_df["smarts"],
            bredt_df["labels"],
            entry_as_inds=True,
        )

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def unstable_graph(severity_threshold: int = 5):
        """Unstable molecular graph to filter out, ideal for generated *de novo* molecules.

        !!! warning
            This method returns problematic patterns and thus patterns with higher severity than the threshold.

        Args:
            severity_threshold: minimum severity for a pattern to **be returned**.
        """
        graph_df = pd.read_csv(get_data_path("graph.csv"))
        # only apply rules with severity >= severity_threshold
        graph_df = graph_df[graph_df["severity"] >= severity_threshold]
        return catalog_from_smarts(
            graph_df["smarts"],
            graph_df["labels"],
            entry_as_inds=True,
        )

    @staticmethod
    @functools.lru_cache(maxsize=32)
    def chemical_groups(filters: Union[str, List[str]] = "medicinal"):
        """Chemical group filter catalogs

        Args:
            filters: list of tag to filter the catalog on.
        """
        chemical_groups = pd.read_csv(get_data_path("chemical_groups.csv"))
        # we cannot keep the nan values in the dataframe
        chemical_groups = chemical_groups.dropna(subset=["smarts"])
        if isinstance(filters, str):
            filters = [filters]
            chemical_groups = chemical_groups[chemical_groups.hierarchy.str.contains("|".join(filters))]

        return catalog_from_smarts(
            chemical_groups["smarts"],
            chemical_groups["name"],
            entry_as_inds=True,
        )



import pytest
import pathlib

import nbformat
from nbconvert.preprocessors.execute import ExecutePreprocessor


ROOT_DIR = pathlib.Path(__file__).parent.resolve()

TUTORIALS_DIR = ROOT_DIR.parent / "docs" / "tutorials"
DISABLE_NOTEBOOKS = []
NOTEBOOK_PATHS = sorted(list(TUTORIALS_DIR.glob("*.ipynb")))
NOTEBOOK_PATHS = list(filter(lambda x: x.name not in DISABLE_NOTEBOOKS, NOTEBOOK_PATHS))

# Discard some notebooks
NOTEBOOKS_TO_DISCARD = ["Basic_Concepts.ipynb"]
NOTEBOOK_PATHS = list(filter(lambda x: x.name not in NOTEBOOKS_TO_DISCARD, NOTEBOOK_PATHS))


@pytest.mark.parametrize("nb_path", NOTEBOOK_PATHS, ids=[str(n.name) for n in NOTEBOOK_PATHS])
def test_notebook(nb_path):
    # Setup and configure the processor to execute the notebook
    ep = ExecutePreprocessor(timeout=600, kernel_name="python3")

    # Open the notebook
    with open(nb_path) as f:
        nb = nbformat.read(f, as_version=nbformat.NO_CONVERT)

    # Execute the notebook
    ep.preprocess(nb, {"metadata": {"path": TUTORIALS_DIR}})



import pytest

import datamol as dm

from medchem.complexity import WhitlockCT
from medchem.complexity import BaroneCT
from medchem.complexity import SMCM
from medchem.complexity import TWC
from medchem.complexity import ComplexityFilter


def test_available_list_cf():
    listed = set(ComplexityFilter.list_default_available_filters())

    assert listed == set(ComplexityFilter.COMPLEXITY_FNS.keys())

    with pytest.raises(ValueError):
        ComplexityFilter(limit="fake")

    with pytest.raises(ValueError):
        # sas is not defined for the zinc_filter
        ComplexityFilter(limit="99", complexity_metric="sas", threshold_stats_file="zinc_12")

    assert len(ComplexityFilter.list_default_percentile()) > 0


def test_whitlockct():
    """Test WhitlockCT"""

    smiles = [
        "CC(=O)O[C@@H]1C2=C(C)[C@H](C[C@@](O)([C@@H](OC(=O)C)[C@@H]3[C@@]4(CO[C@@H]4C[C@H](O)[C@@]3(C)C1=O)OC(C)=O)C2(C)C)OC(=O)[C@H](O)[C@@H](NC(=O)C)C",
        "C1(C)=CC(C)(C)C2(CCC3)CCC(C)C123",
        "O=C1C=CC(=O)C=C1",
    ]

    expected_outputs = [67, 20, 14]
    mols = [dm.to_mol(x) for x in smiles]

    assert [WhitlockCT(m) for m in mols] == expected_outputs

    with pytest.raises(ValueError):
        WhitlockCT(None)


def test_baronect():
    """Test BaroneCT"""
    smiles_non_chiral = [
        "CC1=CCCC1=O",
        "O=C1CCC2(C)C(=O)CCCC2=C1",
        "CC1(C)CCCC2(C)C3CCC(C2O)C13",
    ]
    smiles_chiral = [
        "O=C1CCC2(C)C(=O)CCCC2=C1",
        "O=NCNCCCCc1cccc(OC2=COC=C2)c1",
        "c1ccccc1C(=O)C1=NC=C(C=N1)CCCCCCNC(=O)C(N)N=O",
        "N1CNCC1SCC=CC=C(CC(C)=O)Cc1cccc(C)c1",
    ]
    expected_outputs = [135, 270, 288] + [290, 363, 512, 419]
    mols_non_chiral = [dm.to_mol(x) for x in smiles_non_chiral]
    mols_chiral = [dm.to_mol(x) for x in smiles_chiral]
    outputs = [BaroneCT(mol) for mol in mols_non_chiral] + [BaroneCT(mol, chiral=True) for mol in mols_chiral]

    assert outputs == expected_outputs


def test_smcm():
    """Test SMCM"""
    mols = [
        dm.to_mol(x)
        for x in [
            "CC1=C[C@@H]2[C@]([C@@H](C1=O)O)([C@]3([C@H]([C@H]([C@H]([C@@]34CO4)O2)O)OC(=O)C)C)CO",
            "CN(C)CCOCCOCCN(C)C",
            "OC(=O)CC(C)(O)CC(O)=O",
        ]
    ]
    expected_outputs = [59.376, 20.034, 20.485]
    outputs = [SMCM(mol) for mol in mols]
    assert outputs == expected_outputs


def test_twc():
    """Test TWC"""
    smiles = [
        "CC(C)C(C)C(C)CC(C)C",
        "C[Pt](C)(C)(C)(C)CCCC",
        "CC(C)C(C)CC(C)CCCC",
        "CC(C)CCC(C)C(C)CCC",
        "CCC(C)C(C(C)C)C(C)CC",
        "CCC(C)CC(C)CCCCC",
        "CCC(C)CCCC(CC)CC",
    ]

    mols = [dm.to_mol(x) for x in smiles]
    expected_outputs = [21784, 21784, 40145, 40145, 69926, 31474, 31474]
    outputs = [TWC(mol, log10=False) for mol in mols]
    assert outputs == expected_outputs

    TWC(mols[0], log10=True)


def test_complexity_filter():
    cf = ComplexityFilter(threshold_stats_file="zinc_12")

    smiles = dm.data.cdk2().smiles.iloc[:10].values

    expected_outputs = [False, True, False, False, False, False, True, True, True, False]
    outputs = [cf(dm.to_mol(x)) for x in smiles]

    cf2 = ComplexityFilter(
        limit="99",
        complexity_metric="sas",
        threshold_stats_file="zinc_15_available",
    )
    _ = [cf2(dm.to_mol(x)) for x in smiles]

    cf3 = ComplexityFilter(
        limit="max",
        complexity_metric="clogp",
        threshold_stats_file="zinc_15_available",
    )
    _ = [cf3(dm.to_mol(x)) for x in smiles]

    cf4 = ComplexityFilter(
        limit="999",
        complexity_metric="whitlock",
        threshold_stats_file="zinc_15_available",
    )
    _ = [cf4(dm.to_mol(x)) for x in smiles]

    assert outputs == expected_outputs



from typing import List
from typing import Optional
from typing import Union
from typing import Any
from typing import Sequence

import os
import numpy as np
import datamol as dm

from functools import partial
from loguru import logger

from rdkit.Chem import rdfiltercatalog  # type: ignore

from medchem.structural import CommonAlertsFilters
from medchem.structural import NIBRFilters
from medchem.catalogs import NamedCatalogs
from medchem.catalogs import merge_catalogs
from medchem.complexity import ComplexityFilter
from medchem.groups import ChemicalGroup
from medchem.rules import RuleFilters


def alert_filter(
    mols: Sequence[Union[str, dm.Mol]],
    alerts: List[str],
    alerts_db: Optional[Union[os.PathLike, str]] = None,
    n_jobs: Optional[int] = 1,
    progress: bool = False,
    return_idx: bool = False,
) -> np.ndarray:
    """Filter a dataset of molecules, based on common structural alerts and specific rules.

    ??? tip "True is good"
        Returning `True` means the molecule does not match any of the structural alerts.

    !!! info "See Also"
        `alert_filter` is a convenient functional API for the `medchem.structural.CommonAlertsFilters` class.

    Args:
        mols: List of molecules to filter
        alerts: List of alert collections to screen for. See `CommonAlertsFilters.list_default_available_alerts()`
        alerts_db: Path to the alert file name.
            The internal default file (`alerts.csv`) will be used if not provided
        n_jobs: Number of workers to use
        progress: Whether to show progress bar
        return_idx: Whether to return the filtered index

    Returns:
        filtered_mask: boolean array (or index array) where true means
            the molecule IS OK (not found in the alert catalog).

    """

    custom_filters = CommonAlertsFilters(alerts_set=alerts, alerts_db_path=alerts_db)
    results = custom_filters(mols, n_jobs=n_jobs, progress=progress)

    if return_idx:
        return results.query("pass_filter == True").index.values.astype(int)

    return results["pass_filter"].to_numpy()


def nibr_filter(
    mols: Sequence[Union[str, dm.Mol]],
    n_jobs: Optional[int] = None,
    max_severity: int = 10,
    progress: bool = False,
    return_idx: bool = False,
) -> np.ndarray:
    """
    Filter a set of molecules based on the Novartis Institutes for BioMedical Research screening deck curation process
    [Schuffenhauer, A. et al. Evolution of Novartis' small molecule screening deck design, J. Med. Chem. (2020)](https://dx.doi.org/10.1021/acs.jmedchem.0c01332)


    The severity argument corresponds to the accumulated severity for a compounds accross all pattern in the catalog.

    ??? tip "True is good"
        Returning `True` means the molecule does not match any of the structural alerts.

    !!! info "See Also"
        `nibr_filter` is a convenient functional API for the `medchem.structural.NIBRFilters` class.

    Args:
        mols: list of input molecules
        n_jobs: number of parallel job to run. Sequential by default
        max_severity: maximum severity allowed. Default is <10
        progress: whether to show progress bar
        return_idx: Whether to return the filtered index

    Returns:
        filtered_mask: boolean array (or index array) where true means the molecule
            IS NOT REJECTED (i.e not found in the alert catalog).

    """

    filt_obj = NIBRFilters()
    results = filt_obj(mols, n_jobs=n_jobs, progress=progress)

    mask = results["severity"] < max_severity

    if return_idx:
        return results[mask].index.values.astype(int)

    return mask.to_numpy()


def catalog_filter(
    mols: Sequence[Union[str, dm.Mol]],
    catalogs: List[Union[str, rdfiltercatalog.FilterCatalog]],
    return_idx: bool = False,
    n_jobs: Optional[int] = -1,
    progress: bool = False,
    progress_leave: bool = False,
    scheduler: str = "processes",
    batch_size: int = 100,
) -> np.ndarray:
    """Filter a list of compounds according to a catalog of structural alerts and patterns


    ??? tip "True is good"
        Returning `True` means the molecule does not match any of the structural alerts.

    Args:
        mols: list of input molecules
        catalogs: list of catalogs (name or FilterCatalog)
        return_idx: whether to return index or a boolean mask
        n_jobs: number of parallel job to run. Sequential by default
        progress: whether to show progress bar
        progress_leave: whether to leave the progress bar after completion
        scheduler: joblib scheduler to use
        batch_size: batch size for parallel processing. Note that `batch_size` should be
            increased if the number of used CPUs gets very large.

    Returns:
        filtered_mask: boolean array (or index array) where true means the molecule is not found in the catalog.
    """

    # Build and merge the catalogs
    named_catalogs = []
    for catalog in catalogs:
        if catalog == "bredt":
            raise ValueError(
                "It is not recommended to use the 'bredt' catalog here. Please use the `bredt_filter`."
            )
        if catalog == "nibr":
            raise ValueError(
                "You shouldn't use the nibr catalog here. Please use the `screening_filter` function instead."
            )
        if catalog == "bredt-kekulized":
            # we know the compounds have been kekulized
            catalog = "bredt"
        if isinstance(catalog, str):
            catalog_fn = getattr(NamedCatalogs, catalog, None)
            if catalog_fn is None:
                logger.warning(f"Catalog {catalog} not found. Ignoring.")
            else:
                named_catalogs.append(catalog_fn())
        else:
            named_catalogs.append(catalog)

    if len(named_catalogs) < 1:
        raise ValueError("Please provide at least one catalog !")

    catalog = merge_catalogs(*named_catalogs)

    # Serialize the catalog
    catalog_state = catalog.Serialize()

    def _fn(mols_chunk):
        # Init the catalog from the serialized state
        catalog = rdfiltercatalog.FilterCatalog(catalog_state)

        # To mols
        mols_chunk = [dm.to_mol(m) for m in mols_chunk]

        # Match the mols
        return [catalog.HasMatch(m) for m in mols_chunk]

    # Run the matching
    toxic = dm.parallelized_with_batches(
        _fn,
        mols,
        batch_size=batch_size,
        n_jobs=n_jobs,
        scheduler=scheduler,
        progress=progress,
        tqdm_kwargs=dict(
            desc="Filtering with catalogs",
            leave=progress_leave,
        ),
        flatten_results=True,
    )

    toxic = np.asarray(toxic)
    filtered_idx = [i for i, bad in enumerate(toxic) if not bad]

    if return_idx:
        return np.asarray(filtered_idx)

    return np.bitwise_not(toxic)


def chemical_group_filter(
    mols: Sequence[Union[str, dm.Mol]],
    chemical_group: ChemicalGroup,
    exact_match: bool = False,
    return_idx: bool = False,
    n_jobs: Optional[int] = None,
    progress: bool = False,
    progress_leave: bool = False,
    scheduler: str = "threads",
) -> np.ndarray:
    """Filter a list of compounds according to a chemical group instance.

    !!! warning
        This function will return the list of molecules that **DO NOT** match the chemical group.

    !!! info "See Also"
        Consider exploring the `medchem.groups.ChemicalGroup` class.

    Args:
        mols: list of input molecules
        chemical_group: a chemical group instance with the required functional groups to use.
        exact_match: whether to use an exact match of the chemical group patterns (will switch to smiles )
        return_idx: whether to return index or a boolean mask
        n_jobs: number of parallel job to run. Sequential by default
        progress: whether to show progress bar
        progress_leave: whether to leave the progress bar after completion
        scheduler: joblib scheduler to use

    Returns:
        filtered_mask: boolean array (or index array) where true means the molecule DOES NOT MATCH the groups.
    """

    if isinstance(chemical_group, ChemicalGroup):
        chemical_group = chemical_group.get_catalog(exact_match=exact_match)

    return catalog_filter(
        mols,
        [chemical_group],
        return_idx=return_idx,
        n_jobs=n_jobs,
        progress=progress,
        progress_leave=progress_leave,
        scheduler=scheduler,
    )


def rules_filter(
    mols: Sequence[Union[str, dm.Mol]],
    rules: Union[List[Any], RuleFilters],
    return_idx: bool = False,
    n_jobs: Optional[int] = None,
    progress: bool = False,
    progress_leave: bool = False,
    scheduler: str = "processes",
) -> np.ndarray:
    """Filter a list of compounds according to a predefined set of rules

    ??? tip "True is good"
        Returning `True` means the molecule passes all the rules.

    !!! info "See Also"
        Consider exploring the `medchem.rules.RuleFilters` class.

    Args:
        mols: list of input molecules
        rules: list of rules to apply to the input molecules.
        return_idx: whether to return index or a boolean mask
        n_jobs: number of parallel job to run. Sequential by default
        progress: whether to show progress bar
        scheduler: joblib scheduler to use

    Returns:
        filtered_mask: boolean array (or index array) where true means the molecule MATCH the rule constraints.
    """

    if not isinstance(rules, RuleFilters):
        rules = RuleFilters(rules)

    results = rules(
        mols,
        n_jobs=n_jobs,
        progress=progress,
        progress_leave=progress_leave,
        scheduler=scheduler,
    )

    mask = results["pass_all"].to_numpy()

    if return_idx:
        return results[mask].index.to_numpy()

    return mask


def complexity_filter(
    mols: Sequence[Union[str, dm.Mol]],
    complexity_metric: str = "bertz",
    threshold_stats_file: str = "zinc_15_available",
    limit: str = "99",
    return_idx: bool = False,
    n_jobs: Optional[int] = None,
    progress: bool = False,
    progress_leave: bool = False,
    scheduler: str = "processes",
) -> np.ndarray:
    """Filter a list of compounds according to a complexity metric


    ??? tip "True is good"
        Returning `True` means the molecule passes the complexity filters.

    !!! info "See Also"
        Consider exploring the `medchem.complexity.ComplexityFilter` class.

    Args:
        mols: list of input molecules
        complexity_metric: complexity metric to use
            Use `ComplexityFilter.list_default_available_filters` to list default filters.
            The following complexity metrics are supported by default:

            - `bertz`: bertz complexity index
            - `sas`: synthetic accessibility score  (`zinc_15_available` only)
            - `qed`: qed score  (`zinc_15_available` only)
            - `clogp`: clogp for how greasy a molecule is compared to other in the same mw range  (`zinc_15_available` only)
            - `whitlock`: whitlock complexity index
            - `barone`: barone complexity index
            - `smcm`: synthetic and molecular complexity
            - `twc`:  total walk count complexity  (`zinc_15_available` only)

        threshold_stats_file: complexity threshold statistics file to use
        limit: complexity outlier percentile to use
        return_idx: whether to return index or a boolean mask
        n_jobs: number of parallel job to run. Sequential by default
        progress: whether to show progress bar
        progress_leave: whether to leave the progress bar after completion
        scheduler: joblib scheduler to use

    Returns:
        filtered_mask: boolean array (or index array) where true means the molecule MATCH the rules.
    """

    cf = ComplexityFilter(
        limit=limit,
        complexity_metric=complexity_metric,
        threshold_stats_file=threshold_stats_file,
    )

    not_complex = dm.parallelized(
        cf,
        mols,
        n_jobs=n_jobs,
        scheduler=scheduler,
        progress=progress,
        tqdm_kwargs=dict(
            desc="Complexity Evaluation",
            leave=progress_leave,
        ),
    )
    not_complex = np.asarray(not_complex)
    filtered_idx = [i for i, good in enumerate(not_complex) if good]

    if return_idx:
        return np.asarray(filtered_idx)

    return not_complex


def bredt_filter(
    mols: Sequence[Union[str, dm.Mol]],
    return_idx: bool = False,
    n_jobs: Optional[int] = None,
    progress: bool = False,
    progress_leave: bool = False,
    scheduler: str = "threads",
    batch_size: int = 100,
) -> np.ndarray:
    """Filter a list of compounds according to [Bredt's rules](https://en.wikipedia.org/wiki/Bredt%27s_rule)

    ??? tip "True is good"
        Returning `True` means the molecule does not violate the Bredt's rules.

    Args:
        mols: list of input molecules
        return_idx: whether to return index or a boolean mask
        n_jobs: number of parallel job to run. Sequential by default
        progress: whether to show progress bar
        progress_leave: whether to leave the progress bar after completion
        scheduler: joblib scheduler to use
        batch_size: batch size for parallel processing. Note that `batch_size` should be
            increased if the number of used CPUs gets very large.

    Returns:
        filtered_mask: boolean array (or index array) where true means the molecule is not toxic.
    """

    mols = dm.parallelized(
        partial(dm.to_mol, kekulize=True),
        mols,
        n_jobs=n_jobs,
        progress=progress,
        tqdm_kwargs=dict(desc="To mol", leave=False),
    )

    return catalog_filter(
        mols=mols,
        catalogs=["bredt-kekulized"],  # already kekulized mols
        return_idx=return_idx,
        n_jobs=n_jobs,
        progress=progress,
        progress_leave=progress_leave,
        scheduler=scheduler,
        batch_size=batch_size,
    )


def molecular_graph_filter(
    mols: Sequence[Union[str, dm.Mol]],
    max_severity: Optional[int] = 5,
    return_idx: bool = False,
    n_jobs: Optional[int] = None,
    progress: bool = False,
    progress_leave: bool = False,
    scheduler: str = "threads",
) -> np.ndarray:
    """Filter a list of compounds according to unstable molecular graph patterns.
    This list was obtained from observation around technically valid molecular graphs from
    deep generative models that are not stable.

    The disallowed graphs are:

    -  K3,3 or K2,4 structures
    -  Cone of P4 or K4 with 3-ear
    -  Node in more than one ring of length 3 or 4

    ??? tip "True is good"
        Returning `True` means the molecule does not violate the molecular graph instability rules.

    Args:
        mols: list of input molecules
        max_severity: maximum acceptable severity (1-10). Default is <5
        return_idx: whether to return index or a boolean mask
        n_jobs: number of parallel job to run. Sequential by default
        progress: whether to show progress bar
        progress_leave: whether to leave the progress bar after completion
        scheduler: joblib scheduler to use

    Returns:
        filtered_mask: boolean array (or index array) where true means the molecule is not toxic.
    """

    if max_severity is None:
        max_severity = 5

    catalog = NamedCatalogs.unstable_graph(severity_threshold=max_severity)

    if isinstance(mols[0], str):
        mols = dm.parallelized(
            dm.to_mol,
            mols,
            n_jobs=n_jobs,
            progress=progress,
            tqdm_kwargs=dict(desc="To mol", leave=progress_leave),
        )

    toxic = dm.parallelized(
        catalog.HasMatch,
        mols,
        n_jobs=n_jobs,
        scheduler=scheduler,
        progress=progress,
        tqdm_kwargs=dict(desc="Match", leave=progress_leave),
    )
    toxic = np.asarray(toxic)
    filtered_idx = np.where(~toxic)[0]

    if return_idx:
        return filtered_idx

    return np.bitwise_not(toxic)


def lilly_demerit_filter(
    mols: Sequence[Union[str, dm.Mol]],
    max_demerits: Optional[int] = 160,
    return_idx: bool = False,
    n_jobs: Optional[int] = None,
    progress: bool = False,
    progress_leave: bool = False,
    scheduler: str = "threads",
    batch_size: int = 5_000,
    **kwargs: Any,
) -> np.ndarray:
    """Run the Eli Lilly's demerit filter on current list of molecules

    ??? tip "True is good"
        Returning `True` means the molecule does not violate the demerit rules.

    !!! info "See Also"
        Consider exploring the `LillyDemeritsFilters` class in `medchem.structural.lilly_demerits`

    Args:
        mols: list of input molecules as smiles preferably
        max_demerits: Cutoff to reject molecules Defaults to 160.
        return_idx: whether to return a mask or a list of valid indexes
        progress: whether to show progress bar
        progress_leave: whether to leave the progress bar after completion
        scheduler: joblib scheduler to usescheduler
        batch_size: batch size for parallel processing.
        kwargs: parameters specific to the `demerits.score` function

    Returns:
        filtered_mask: boolean array (or index array) where true means the molecule is ok.
    """

    from medchem.structural.lilly_demerits import LillyDemeritsFilters

    dfilter = LillyDemeritsFilters()

    results = dfilter(
        mols,
        n_jobs=n_jobs,
        progress=progress,
        progress_leave=progress_leave,
        scheduler=scheduler,
        batch_size=batch_size,
        **kwargs,
    )

    results = results[
        (results["pass_filter"] == True)  # noqa: E712
        & ((results["demerit_score"].isna()) | (results["demerit_score"] < max_demerits))
    ]

    filtered_idx = results.index.values.astype(int)
    filtered_mask = np.zeros(len(mols), dtype=bool)
    filtered_mask[filtered_idx] = True

    if return_idx:
        return filtered_idx

    return filtered_mask


def protecting_groups_filter(
    mols: Sequence[Union[str, dm.Mol]],
    return_idx: bool = False,
    protecting_groups: List[str] = [
        "fmoc",
        "tert-butoxymethyl",
        "tert-butyl carbamate",
        "tert-butyloxycarbonyl",
    ],
    n_jobs: Optional[int] = None,
    progress: bool = False,
    progress_leave: bool = False,
    scheduler: str = "threads",
) -> np.ndarray:
    """Filter a list of compounds according to match to known protecting groups.


    !!! warning
         This function will return the list of molecules that **DO NOT** have the protecting groups.

     !!! info "See Also"
         This is a syntaxic sugar for calling chemical_group_filter with the protecting groups subset.

     Args:
         mols: list of input molecules
         protecting_groups: type of protection group to consider if not provided, will use all **(not advised)**
         return_idx: whether to return index or a boolean mask
         n_jobs: number of parallel job to run. Sequential by default
         progress: whether to show progress bar
         progress_leave: whether to leave the progress bar after completion
         scheduler: joblib scheduler to use

     Returns:
         filtered_mask: boolean array (or index array) where true means the molecule DOES NOT MATCH the groups.
    """

    chemical_group = ChemicalGroup("protecting_groups")
    if protecting_groups and len(protecting_groups) > 0:
        chemical_group = chemical_group.filter(protecting_groups)

    protected = dm.parallelized(
        partial(chemical_group.has_match, exact_match=True, terminal_only=True),
        mols,
        n_jobs=n_jobs,
        progress=progress,
        scheduler=scheduler,
        tqdm_kwargs={"desc": "Checking protecting groups", "leave": progress_leave},
    )
    filtered_idx = [i for i, bad in enumerate(protected) if not bad]

    if return_idx:
        return np.asarray(filtered_idx)

    return np.bitwise_not(protected)
